package org.processmining.datadiscovery;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;

import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.datadiscovery.model.DecisionPointResult;
import org.processmining.datadiscovery.model.DiscoveredDataPetriNet;
import org.processmining.datadiscovery.plugins.DecisionMining;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.util.Pair;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.InductiveMiner.mining.MiningParameters;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIMf;
import org.processmining.plugins.InductiveMiner.plugins.IMPetriNet;
import org.processmining.pmind.help.PMInDHelp;
import org.processmining.pmind.logs.EventLogOperations;

public class QuickAndDirtyPlugin {

	@Plugin(name = "_Quick and dirty", parameterLabels = {}, returnLabels = { "Petri Net with Data", "Initial Marking",
	"Final Marking" }, returnTypes = { DiscoveredDataPetriNet.class, Marking.class,
			Marking.class }, userAccessible = false, help = PMInDHelp.TEXT)

	@UITopiaVariant(affiliation = "University of Edinburgh", author = "Johannes De Smedt", email = "johannes.desmedt@ed.ac.uk")

	public Object[] run(UIPluginContext context) throws Exception {

		
		//XLog log = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/17Mini30.xes");
		XLog log = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/17Mini10.xes");
		log = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/Sepsis.xes");
		//log = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/examp.mxml");
		
		//log = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/BPI_Challenge_2012.xes");
		//log = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/RTFMP.xes");
		log = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/BPI Challenge 2017.xes");

	
		
		HashMap<Boolean,HashSet<Pair<Integer,Double>>> result = new HashMap<Boolean,HashSet<Pair<Integer,Double>>>();
		for(Boolean b: new Boolean[]{false,true})
			if(b){
				HashSet<Pair<Integer,Double>> s = new HashSet<Pair<Integer,Double>>();
				result.put(b, s);
				for(Integer i: new Integer[]{0,1,2,5})
					result.get(b).add(new Pair<Integer,Double>(i,doTheStuff(context, log, b, i)));				
			}else{
				HashSet<Pair<Integer,Double>> s = new HashSet<Pair<Integer,Double>>();
				s.add(new Pair<Integer,Double>(1,doTheStuff(context, log, b, 0)));
				result.put(b, s);
			}
		
		for(Boolean b: result.keySet())
			System.out.println("AUC with "+b+" autocorrelations: "+result.get(b));
		
		return null;
	}
	
	public static double doTheStuff(UIPluginContext context, XLog log, boolean addAuto, int autoCorr) throws Exception{	
		if(addAuto)
			EventLogOperations.addAttributePositionInformationToLog(log, true, autoCorr);
		
		// Mine models with decision tree miner
		DecisionMining dt = new DecisionMining();
		MiningParameters parameters = new MiningParametersIMf();
		parameters.setClassifier(MiningParameters.getDefaultClassifier());
		parameters.setNoiseThreshold(0.0f);

		IMPetriNet im = new IMPetriNet();
		Object[] imO = im.minePetriNet(context, log, parameters);
		Collection<Marking> markings = new HashSet<Marking>();
		markings.add((Marking) imO[2]);

		Object[] o = dt.decisionMiner(context, (Petrinet) imO[0], log);
		
		double auc = 0.0;
		int noCount = 0;
		
		DiscoveredDataPetriNet ddpn = (DiscoveredDataPetriNet) o[0];
		for(Place p: ddpn.getPlaces())
			if(!(ddpn.getDecisionPointResult(p)==null)){
				DecisionPointResult dpr = ddpn.getDecisionPointResult(p);
				System.out.println("AUC: "+addAuto+"  "+dpr.getQualityMeasure());
				auc= auc + dpr.getQualityMeasure();
				noCount++;
				for(Object oo: dpr.getEstimatedGuards().keySet()){
					System.out.println(oo+" - "+dpr.getEstimatedGuards().get(oo).getQualityMeasure());
					System.out.println(oo+" - "+dpr.getEstimatedGuards().get(oo).getExpression());
				}
			}
		return auc/noCount;
	}

}
