package org.processmining.pmind.algorithms;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.impl.XLogImpl;
import org.processmining.framework.packages.PackageManager.Canceller;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.InductiveMiner.mining.MiningParameters;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIMf;
import org.processmining.plugins.InductiveMiner.plugins.IMPetriNet;
import org.processmining.plugins.graphviz.dot.Dot;
import org.processmining.pmind.algorithms.autocorrelations.HelpFunctions;
import org.processmining.pmind.algorithms.concept.AttributeShiftSequence;
import org.processmining.pmind.algorithms.concept.PossibleModel;
import org.processmining.pmind.algorithms.concept.PredictiveModel;
import org.processmining.pmind.algorithms.concept.Shift;
import org.processmining.pmind.algorithms.concept.TraceCluster;
import org.processmining.pmind.models.Activity;
import org.processmining.pmind.models.DMNActivityAttributePair;
import org.processmining.pmind.models.DMNAttribute;
import org.processmining.pmind.models.DMNModel;
import org.processmining.pmind.models.DMNNode;
import org.processmining.pmind.parameters.PMInDParameters;

import com.google.common.collect.Sets;

import weka.attributeSelection.ReliefFAttributeEval;
import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.classifiers.trees.J48;
import weka.classifiers.trees.REPTree;
import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;
import weka.filters.Filter;
import weka.filters.unsupervised.attribute.Normalize;

public abstract class AbstractDMNMiner {

	protected double shiftRatio;
	protected double minSup;
	protected double minTraceProportion;
	protected int minTrace;
	protected double minCorr;
	protected double minDev;
	protected float im;
	protected boolean modelBased;

	protected XLogInfo xLogInfo;
	protected XLog log;
	protected XEventClassifier eventClassifier = MiningParameters.getDefaultClassifier();

	protected Collection<String> attributes;
	protected HashMap<XEventClass, Activity> activities = new HashMap<XEventClass, Activity>();
	protected HashMap<Activity, XEventClass> decisionActivityClasses = new HashMap<Activity, XEventClass>();
	protected Set<DMNAttribute> dmnAttributes = new HashSet<DMNAttribute>();

	protected ArrayList<Attribute> wekaAttributes = new ArrayList<Attribute>();
	protected HashMap<String,ArrayList<String>> data = new HashMap<String, ArrayList<String>>();
	protected HashMap<Integer,DMNActivityAttributePair> daapToAt = new HashMap<Integer,DMNActivityAttributePair>();
	
	protected DMNModel mainModel;
	protected Set<DMNModel> topModels;
	protected DMNAttribute mainAtt;
	protected Set<DMNModel> models;
	protected Collection<TraceCluster> traceClusters = new HashSet<TraceCluster>();

	protected String modelPrint = "";
	
	// Output dot objects
	protected List<Dot> dots = new ArrayList<Dot>();

	public AbstractDMNMiner(XLog log, PMInDParameters parameters) throws Exception{
			this.shiftRatio = (double) parameters.getShiftRatio()/(double) 100;
			this.minSup = (double) parameters.getMinSup()/(double) 100;
			this.minTraceProportion = (double) parameters.getMinTraceProp()/(double) 100;
			this.minDev = (double) parameters.getMinDeviation()/(double) 100;
			this.minCorr = (double) parameters.getMinCorrelation()/(double) 100;
			this.im = (float) parameters.getIMSetting()/(float) 100;
			this.modelBased = parameters.isModelBased();			
			
			xLogInfo = XLogInfoFactory.createLogInfo(log);
			this.log = log;			
			
			attributes = xLogInfo.getEventAttributeInfo().getAttributeKeys();
			for(String attribute: attributes)
				dmnAttributes.add(new DMNAttribute(attribute));			
			
			for(XEventClass eL: xLogInfo.getEventClasses(eventClassifier).getClasses()){			
				Activity d = new Activity(eL, dmnAttributes, false, parameters.isPrintModels());
				activities.put(eL,d);
				decisionActivityClasses.put(d, eL);
			}					
					
			HashSet<Shift> shifts = new HashSet<Shift>();
			// Loop log
			for(int t=0;t<log.size();t++){
				int tp = 0;
				//XTrace newT = new XTraceImpl(log.get(t).getAttributes());
				//newT.getAttributes().put("concept:name", log.get(t).getAttributes().get("concept:name"));
				
				// Loop traces
				for(XEvent e: log.get(t)){
					//System.out.println(e.getAttributes().get("lifecycle:transition").toString());
					if(e.getAttributes().get("lifecycle:transition").toString().equals("start")){		
						XEventClass ec = xLogInfo.getEventClasses(eventClassifier).getClassOf(e);
						XEvent previousEvent = null;
						if(tp!=0)
							previousEvent = log.get(t).get(tp-1);
						Collection<Shift> shiftsForE = activities.get(ec).addEvent(e, previousEvent, e, t, tp, xLogInfo);
						shifts.addAll(shiftsForE);									
					}
					tp++;	
				}
			}					
			System.out.println("Shifts found");
			System.out.println("logsize "+ log.size());			
			System.out.println("Finished");
		}

	public String getAttributeValAtTracePos(DMNAttribute a, int t, int p) {
		return log.get(t).get(p).getAttributes().get(a.toString()).toString();
	}

	public void mineDMNModels() {
		for (XEventClass e : activities.keySet()) {
			Activity da = activities.get(e);

			// Determines whether attributes are input attributes
			// Determines whether attributes are shifting and stores them for future analysis
			da.analyzeAttributes(shiftRatio);
		}

		// Connect all activities based on the attributes that are set or inputed
		// i.e. all set -> input relationships (bidirectional)
		// TODO: look into whether all variables are input variables
		// The model classifies the nodes based on their inputs/outputs as well

		//mainModel = connectActivities();		

		minTrace = (int) (log.size()*minTraceProportion);
		System.out.println("MinTrace: "+minTrace);		
		models = new HashSet<DMNModel>();
		topModels = new HashSet<DMNModel>(); 
		
		buildModelForActivities();

		//Derive trace clusters
		Collection<Collection<Integer>> tcs = new HashSet<Collection<Integer>>();
		Set<DMNModel> toKeep = new HashSet<DMNModel>();
		for (DMNModel m : models) {
			if (m.getEdges().size() > 1) {
				System.out.println(m.getLabel() + " has a trace cluster of size " + m.getTraces().size());
				tcs.add(m.getTraces());
				toKeep.add(m);
				if(m.getSuperModel()==null)
					topModels.add(m);
			}
		}
		models.retainAll(toKeep);
		System.out.println("#models: " + models.size());

		for (Collection<Integer> t : tcs) {
			System.out.println("Trace cluster: " + t.size());

			XLog l = new XLogImpl(log.getAttributes());
			for (Integer i : t)
				l.add(log.get(i));

			MiningParameters parameters = new MiningParametersIMf();
			//parameters.setClassifier(xLogInfo.getEventClasses().getClassifier());
			parameters.setNoiseThreshold(im);
			parameters.setLifeCycleClassifier(MiningParameters.getDefaultLifeCycleClassifier());

			IMPetriNet im = new IMPetriNet();
			Object[] imO = null;
			if(this.im<=100){
				imO = im.minePetriNet(l, parameters, new Canceller() {
					public boolean isCancelled() {
						return false;
					}
				});
			}
			Collection<Marking> markings = new HashSet<Marking>();
			markings.add((Marking) imO[2]);

			TraceCluster tc = new TraceCluster(new HashSet<DMNModel>(), t, imO);

			for (DMNModel m : models) {
				if (m.getTraces().equals(t)) {
					tc.getModels().add(m);
					m.addPetrinet(imO);
				}
			}
			traceClusters.add(tc);
		}
	}

	protected abstract void buildModelForActivities();

	protected Set<DMNModel> eliminateDuplicates(Set<DMNModel> org) {
		//Set<DMNModel> newC = new HashSet<DMNModel>();
		Set<DMNModel> toR = new HashSet<DMNModel>();
		for (DMNModel m : org)
			for (DMNModel c : org)
				if (!m.equals(c) && c.contains(m)) {
					toR.add(m);
					if (m.getSuperModel() != null) {
						m.getSuperModel().getSubModels().remove(m);
						for (DMNModel ms : m.getSubModels())
							ms.setSuperModel(m.getSuperModel());
					} else {
						for (DMNModel ms : m.getSubModels())
							ms.setSuperModel(null);
					}
					break;
				}
		org.removeAll(toR);
		return org;
	}

	protected abstract void findRelatedNodes(int level, DMNModel newModel, AttributeShiftSequence assD,
			DMNActivityAttributePair topDAP, Set<Integer> shiftingTraces) ;

	public void addNodes(int level, DMNModel m, DMNActivityAttributePair topDAP, PossibleModel p) {
		for (DMNActivityAttributePair daap : p.getDAAPS()) {			
			// Revise correlation:			
			PossibleModel pm = new PossibleModel(topDAP, daap, m.getTraces());		
			double corr = buildModel(pm,true).getOutcome();
			
			System.out.println("Corr: "+corr);
			
			if(corr>minCorr){
				DMNNode n2 = m.addNode(daap, "#8fffff");
				DMNNode n3 = m.addNode(topDAP);
				m.addEdge(n2, n3, p.getTraces().size(), Math.round(corr*100.0)/100.0);//daap.getCorr());
			}
		}
		if(modelBased && !(topDAP==null)){
			System.out.println("Adding knowledge node: "+topDAP);
			m.addKnowledgeNode(modelPrint, topDAP);
		}
		for (DMNActivityAttributePair daap : p.getDAAPS()) {
			String space = "";
			for (int i = 1; i < level; i++)
				space += "|_";

			System.out.println(space + "Going down rabbit hole for " + daap);
			findRelatedNodes(level + 1, m,
					findASS(daap.getTraces(), daap.getShift(), daap.getActivity(), daap.getAttribute()), daap,
					p.getTraces());
			System.out.println(space+"done with "+daap);
		}
		System.out.println(p + " is dealt with");
	}

	private AttributeShiftSequence findASS(Collection<Integer> shiftingTraces, int noShift, Activity a,
			DMNAttribute o) {
		Collection<Shift> shiftsForD = new HashSet<Shift>();

		Collection<Integer> subST = new HashSet<Integer>();
		for (Integer i : shiftingTraces) {
			if (a.getShifts(o, i).size() >= noShift /*&& (noShift-1)>0*/) {
				subST.add(i);
				shiftsForD.add(a.getShifts(o, i).get(noShift - 1));
			}
		}
		return new AttributeShiftSequence(shiftsForD);
	}

	protected Set<PossibleModel> findPossibleModels(DMNActivityAttributePair topDAAP, Set<PossibleModel> possibleModels,
			Collection<DMNActivityAttributePair> pairs) {
		// Check all pairs
		//TODO: check the shift number: should be different tuple/pair
		for (DMNActivityAttributePair aap : pairs) {
			//System.out.println("Checking: "+aap.toString());
			Set<PossibleModel> toAdd = new HashSet<PossibleModel>();

			for (PossibleModel pmP : possibleModels) {
				// Possible model does not contain activity-attribute-pair
				if (!pmP.getDAAPS().contains(aap)) {
					Set<Integer> inter = Sets.intersection(pmP.getTraces(), aap.getTraces());

					// If the AA-pair uses the exact same traces
					if (pmP.getTraces().equals(aap.getTraces())) {
						//System.out.println(aap + " overlaps in: " + pmP);
						pmP.addDAAP(aap);

						// The AA-pair is contained within -> create new possible model for new collection of AAs	
					} else if (pmP.getTraces().containsAll(aap.getTraces())) {
						//System.out.println(aap + " contained in: " + pmP + " " + aap.getTraces());
						PossibleModel pmN = new PossibleModel(pmP.getDAAPS(), topDAAP, aap.getTraces());
						pmN.addDAAP(aap);
						if (aap.getTraces().size() > minTrace)
							toAdd.add(pmN);

						// Some of the traces of AA-pair included -> add to current model (will be used only for these traces)
					} else if (inter != null) {
						//System.out.println(aap + " intersects with: " + pmP);
						// If intersection is big enough
						if (inter.size() >= pmP.getTraces().size() * minDev) {
							pmP.addDAAP(aap);
							pmP.setTraces(inter);

							// Else create new possible model
						} else {
							PossibleModel pmN = new PossibleModel(pmP.getDAAPS(), topDAAP, inter);
							if (aap.getTraces().size() > minTrace)
								pmN.addDAAP(aap);
							toAdd.add(pmN);
						}
					}
				}
			}
			possibleModels.addAll(toAdd);
			Set<Integer> notCovered = notCoveredByModels(possibleModels, aap.getTraces());
			// Traces that are not covered by other models but by this AA -> if enough traces, create new model
			if (notCovered.size() > minTrace) {
				PossibleModel pmN = new PossibleModel(aap, topDAAP, notCovered);
				possibleModels.add(pmN);
			}
		}
		return possibleModels;
	}

	private Set<Integer> notCoveredByModels(Collection<PossibleModel> pms, Collection<Integer> traces) {
		Set<Integer> notCovered = new HashSet<Integer>();
		for (Integer i : traces) {
			boolean contained = false;
			for (PossibleModel pm : pms)
				if (pm.getTraces().contains(i)) {
					contained = true;
					break;
				}
			if (!contained)
				notCovered.add(i);
		}
		return notCovered;
	}
	
	public PredictiveModel buildModel(PossibleModel m, boolean doig) {						
		Instances trainingSet = createTrainingSet(m);		
		if(trainingSet.isEmpty())
			new PredictiveModel(null, 0.0);
		int classIndex = wekaAttributes.size()-1;

		//If the attributes have more than two values, and the trainingset is larger than 1, construct model
		if (!(trainingSet.numDistinctValues(classIndex) == 1) && trainingSet.size() > 10) {
			//DMNDecisionTable dt = new DMNDecisionTable();
			//dt.setDisplayRules(true);	
			//dt.buildClassifier(trainingSet);					

			// Only build model for correlation
			if (doig) {
				try {
					boolean num = false;
					if (!num) {
						ReliefFAttributeEval ige = new ReliefFAttributeEval();
						ige.buildEvaluator(trainingSet);
						//int removed = 0;
						for (int wa = 0; wa < wekaAttributes.size() - 1; wa++) {
							double ig = ige.evaluateAttribute(wa);
							return new PredictiveModel(null, ig);
							//	if(ig<1){							
							//		trainingSet.deleteAttributeAt(wa-removed);
							//		m.getDAAPS().remove(daapToAt.get(wa));
							//		removed++;
							//	}
						}
					}
				} catch (Exception e) {
					return new PredictiveModel(null, 0.0);
				}
			}

			System.out.println("Building a weka model for "+ m.getTopDAAP().getAttribute() +" with " 
						+ trainingSet.size() + " and " + wekaAttributes.size()
						+ " attributes "+m.getDAAPS());
			
			try {
				Classifier rf = new J48();
				double outcome = 0.0;
				boolean num = trainingSet.attribute(classIndex).isNumeric();
				
//				System.out.println("#wekaAttributes: "+wekaAttributes.size());
//				for(Attribute a: wekaAttributes){
//					System.out.println(a.name()+" "+a.numValues());
//				}
				
				if (num) {
					//System.out.println("REPT");
					rf = new REPTree();
				}
				Evaluation ev = null;				
				try {
					rf.buildClassifier(trainingSet);
					ev = new Evaluation(trainingSet);
					//System.out.println(rf.toString());
					ev.crossValidateModel(rf, trainingSet, 5, new Random(1));
					outcome = 1 - ev.rootMeanSquaredError();
				} catch (Exception e) {
					e.printStackTrace();
				}

				modelPrint = rf.toString();
				System.out.println(outcome);
				if (outcome > minSup) {
					return new PredictiveModel(rf, outcome);
				}
			} catch (Exception e) {
				return new PredictiveModel(null, 0.0);
			}
		}
		return new PredictiveModel(null, 0.0);
	}	
	
	public Instances createTrainingSet(PossibleModel m){		
		DMNActivityAttributePair target = m.getTopDAAP();
		wekaAttributes = new ArrayList<Attribute>();
		data = new HashMap<String, ArrayList<String>>();
		daapToAt = new HashMap<Integer,DMNActivityAttributePair>();
				
		// Collecting the data	
		int at = 0;
		for(DMNActivityAttributePair daap: m.getDAAPS()){
			expandDataset(daap, m.getTraces());		
			daapToAt.put(at,daap);
			at++;
		}
		expandDataset(target, m.getTraces());
		daapToAt.put(at,target);	

		int size = data.values().iterator().next().size();
				
		Instances trainingSet = new Instances("", wekaAttributes, size);
		// Loop all instances
		for (int i = 0; i < size; i++) {
			Instance instance = new DenseInstance(data.keySet().size());
			// Loop all attributes
			for (int att = 0; att < wekaAttributes.size(); att++) {
				Attribute weA = wekaAttributes.get(att);
				if (weA.isNumeric())
					instance.setValue(weA, Double.valueOf(data.get(weA.name()).get(i)));
				else
					instance.setValue(weA, data.get(weA.name()).get(i));
			}
			trainingSet.add(instance);
		}
		int classIndex = wekaAttributes.size()-1;
		
		if(wekaAttributes.size()<2)
			return new Instances("", wekaAttributes, 0);
		
		// Normalize numeric values
		Normalize d = new Normalize();
		try {
			d.setInputFormat(trainingSet);
			d.setIgnoreClass(true);
			trainingSet = Filter.useFilter(trainingSet, d);
		} catch (Exception e) {
			e.printStackTrace();
		}		
		trainingSet.setClassIndex(classIndex);
		
		return trainingSet;
	}
	
	public void expandDataset(DMNActivityAttributePair daap, Set<Integer> traces){
		String a = daap.toString();
		ArrayList<String> subset = new ArrayList<String>();
		
		for(Integer i: traces)
			subset.add(daap.getValueOfTrace(i));
		data.put(a, subset);
		
		Attribute wekaAttribute = null;
		Collection<String> levels = HelpFunctions.getLevels(subset);
		if(levels.size() > (traces.size()*0.5)){
			System.out.println("Not using "+ daap);
		} else if (levels.size() > 1) {
			ArrayList<String> classVal = new ArrayList<String>(levels.size());
			classVal.addAll(levels);
			wekaAttribute = new Attribute(a, classVal);
			wekaAttributes.add(wekaAttribute);
			//System.out.println(daap+"  is not numeric");
		} else if (levels.size() == 0) {
			wekaAttribute = new Attribute(a);
			wekaAttributes.add(wekaAttribute);
			//System.out.println(daap+"  is numeric");
		}
	}	

	public Set<DMNModel> getTopModels(){
		return topModels;
	}
	
	public Collection<TraceCluster> getTraceClusters() {
		return traceClusters;
	}

	public XLogInfo getInfo() {
		return xLogInfo;
	}

	public XLog getLog() {
		return log;
	}

	public Set<DMNModel> getModels() {
		return models;
	}

}
