package org.processmining.pmind.algorithms;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.model.XLog;
import org.processmining.pmind.algorithms.concept.AttributeShiftSequence;
import org.processmining.pmind.algorithms.concept.PossibleModel;
import org.processmining.pmind.algorithms.concept.Shift;
import org.processmining.pmind.models.Activity;
import org.processmining.pmind.models.DMNActivityAttributePair;
import org.processmining.pmind.models.DMNAttribute;
import org.processmining.pmind.models.DMNModel;
import org.processmining.pmind.parameters.PMInDParameters;

import com.google.common.collect.Lists;

public class PMInDNoShifts extends AbstractDMNMiner{

	public PMInDNoShifts(XLog log, PMInDParameters parameters) throws Exception{
		super(log, parameters);
	}
						
	protected void buildModelForActivities(){		
		for (XEventClass e : activities.keySet()) {
			Activity topActivity = activities.get(e);

			for (DMNAttribute o : topActivity.getSettedAttributes()) {
				mainAtt = o;
				// Traces in which o is shifted by d
				Set<Integer> shiftingTraces = topActivity.getShiftTraces(o);
				Map<Integer,String> values = new HashMap<Integer,String>();
				AttributeShiftSequence assD = new AttributeShiftSequence(topActivity.getShifts(o));
				
				for(Integer i: shiftingTraces)
					values.put(i, "");

				System.out.println("------------------------------------------------------");
				System.out.println(
						topActivity + " setting " + o.toString() + " " + shiftingTraces.size());
				System.out.println("------------------------------------------------------");
				
				// Create new model for variable that is set by dn
				DMNActivityAttributePair dap = new DMNActivityAttributePair(topActivity, o, values, 0, 1);
				
				DMNModel newModel = new DMNModel(topActivity + "->" + o.toDotString() + "_" + shiftingTraces.size(),
						shiftingTraces, dap);
								
				// Train connection between setted attribute and the activity's other activities
				models.add(newModel);
				findRelatedNodes(1, newModel, assD, dap, shiftingTraces);

				// Rude model model is you big enough?
				if (newModel.getEdges().size() > 1) {
					System.out.println(
							"Making model for " + topActivity + " deciding " + o.toString() + " for shift " + 1);
				}
			}
		}
		Set<DMNModel> ms = eliminateDuplicates(models);
		models = ms;
		//System.out.println("Models final: "+ms.size());
		for(DMNModel m: ms){
			if(m.getEdges().size()>1){
				dots.add(m.toDot());
			}
		}
	}
	
	protected void findRelatedNodes(int level, DMNModel newModel, AttributeShiftSequence assD, 
			DMNActivityAttributePair topDAP, Set<Integer> shiftingTraces){
		boolean f = false;
		
		// Check if model already built
//		for(Pair<DMNActivity,DMNAttribute> s: newModel.getTrainedModels())
//			if(s.getFirst().equals(topDAP.getActivity()) && s.getSecond().equals(topDAP.getAttribute())){
//				f=true;
//				break;
//			}
				
		String space = "";
		for(int i=1;i<level;i++)
			space+= "|_";
		
		// Model for input-output already constructed
		if(!f && topDAP.getTraces().size()>minTrace){						
			
			List<DMNActivityAttributePair> pairs = new ArrayList<DMNActivityAttributePair>();

			// For all incoming attributes da set by activity dda
			for (XEventClass e : activities.keySet()) {
				for (DMNAttribute da : topDAP.getActivity().getInputAttributes()) {				
					if (activities.get(e).getSettedAttributes().contains(da)) {
						Activity dda = activities.get(e);
						
						if(da.equals(topDAP.getAttribute()) && dda.equals(topDAP.getActivity()))
							continue;
						
						// The ass of the other activity + attribute
						AttributeShiftSequence ass = new AttributeShiftSequence(dda.getShifts(da));

						Map<Integer,String> valuesUp = new HashMap<Integer,String>();
						Map<Integer,String> valuesLow = new HashMap<Integer,String>();

						Set<Integer> withHeldTraces = new HashSet<Integer>();						
						Set<Integer> instances = new HashSet<Integer>();
						
						for (Integer t : shiftingTraces) {
							ArrayList<Shift> upperAss = assD.getOrderedShiftsOfTrace(t);

							if (upperAss.size() == 0)
								continue;

							ArrayList<Shift> lowerAss = ass.getOrderedShiftsOfTraceBefore(t,
									upperAss.get(upperAss.size() - 1).getPosition());
							int pos = lowerAss.size() - 1;
							for (Shift s : assD.getOrderedShiftsOfTrace(t)) {
								if (pos < s.getPosition() && pos > -1) {
									valuesUp.put(instances.size(),getAttributeValAtTracePos(topDAP.getAttribute(), t, s.getPosition()));
									valuesLow.put(instances.size(),getAttributeValAtTracePos(da, t, lowerAss.get(pos).getPosition()));
									instances.add(instances.size());							
									pos--;
									withHeldTraces.add(t);
								}
							}
						}
						
						Map<Integer,String> values = new HashMap<Integer,String>();
						
						for(Integer tt: withHeldTraces)
							values.put(tt, "");
						
						DMNActivityAttributePair aap = new DMNActivityAttributePair(topDAP.getActivity(), topDAP.getAttribute(), valuesUp, 0,
								0);
						DMNActivityAttributePair aap2 = new DMNActivityAttributePair(dda, da, valuesLow, 0,
								0);
						
						PossibleModel m = new PossibleModel(aap,aap2,instances);
						double corr = buildModel(m,true).getOutcome();

						System.out.println(space + dda + " sets " + da + " , corr with " + topDAP.getAttribute()
						+ " is " + corr + " for " + valuesUp.size() + " and " + valuesLow.size());
						if (corr > minCorr && valuesUp.values().size() > 1 && valuesLow.values().size() > 1
								&& withHeldTraces.size() > minTrace) {

							DMNActivityAttributePair aap3 = new DMNActivityAttributePair(dda, da, values, 1,
									Math.round(corr * 100.0) / 100.0);
							pairs.add(aap3);
						}
					}
				}
			}
			// Sort pairs according to the size of traces supporting them
			Collections.sort(pairs);
					
			Set<PossibleModel> pm = 
					findPossibleModels(topDAP, new HashSet<PossibleModel>(), pairs);
			
			Set<PossibleModel> toRemove = new HashSet<PossibleModel>();
			for(PossibleModel p: pm)
				for(PossibleModel p2: pm)
					if(!p.equals(p2))
						if(p.contains(p2))
							toRemove.add(p2);	
			pm.removeAll(toRemove);			
			
			List<PossibleModel> list = Lists.newArrayList(pm);
			Collections.sort(list);
			for(PossibleModel pms: list)
				System.out.println("Possible model "+pms);
			
			for(PossibleModel p: list){
				if(p.getTraces().size()>minTrace){
					if(p.getTraces().equals(newModel.getTraces())){
						
			//PossibleModel p = new PossibleModel(new HashSet<DMNActivityAttributePair>())			
						System.out.println("\n--------------------------------\nExtending model with "+p.getDAAPS());
						addNodes(level, newModel, topDAP, p);						
					}else{					
						DMNModel m = newModel.clone(newModel.getLabel()+"_"+topDAP.getTraces().size(), p.getTraces());
						models.add(m);				
						newModel.addSubModel(m);
						
						System.out.println("\n--------------------------------\nExtending model with new model for "+p.getDAAPS());
						
						//What does this thing do??
						m.setSuperModel(newModel);
						m.setLabel(newModel.getLabel()+"_|_"+p.getTraces().size());						
						addNodes(level, m, topDAP, p);	
					}
				}		
			}	
		}
	}
	
}
