package org.processmining.pmind.algorithms;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.model.XLog;
import org.processmining.framework.util.Pair;
import org.processmining.pmind.algorithms.concept.AttributeShiftSequence;
import org.processmining.pmind.algorithms.concept.PossibleModel;
import org.processmining.pmind.algorithms.concept.PredictiveModel;
import org.processmining.pmind.algorithms.concept.Shift;
import org.processmining.pmind.models.Activity;
import org.processmining.pmind.models.DMNActivity;
import org.processmining.pmind.models.DMNActivityAttributePair;
import org.processmining.pmind.models.DMNAttribute;
import org.processmining.pmind.models.DMNModel;
import org.processmining.pmind.parameters.PMInDParameters;

import com.google.common.collect.Lists;

public class PMInDShiftBased extends AbstractDMNMiner{
			
	public PMInDShiftBased(XLog log, PMInDParameters parameters) throws Exception{
		super(log, parameters);
	}
				
	protected void buildModelForActivities(){

		for (XEventClass e : activities.keySet()) {
			Activity topActivity = activities.get(e);

			for (DMNAttribute o : topActivity.getSettedAttributes()) {
				mainAtt = o;
				// Traces in which o is shifted by d
				Collection<Integer> shiftingTraces = topActivity.getShiftTraces(o);

				HashMap<Integer, Integer> noShiftsDist = new HashMap<Integer, Integer>();
				// Trying shifting stuff
				// I.e.: do the analysis for a particular shift, not for all shifts
				for (Integer i : shiftingTraces) {
					// shift of attribute o in trace i
					int no = topActivity.getShifts(o, i).size();
					if (!noShiftsDist.containsKey(no))
						noShiftsDist.put(no, 0);
					noShiftsDist.put(no, noShiftsDist.get(no) + 1);
				}

				// noShift stands for the Nth (noShift) shift the pair activity-attribute has
				for (Integer noShift : noShiftsDist.keySet()) {
					// Retain only traces that have enough shifts 
					// E.g. 4 shifts in case 4 shifts are analysed
					Collection<Shift> shiftsForD = new HashSet<Shift>();
					Map<Integer,String> values = new HashMap<Integer,String>();
					
					Set<Integer> subST = new HashSet<Integer>();
					for (Integer i : shiftingTraces) {
						if (topActivity.getShifts(o, i).size() >= noShift) {
							subST.add(i);
							shiftsForD.add(topActivity.getShifts(o, i).get(noShift - 1));
							values.put(i,
									getAttributeValAtTracePos(o, i, topActivity.getShifts(o, i).get(noShift - 1).getPosition()));
						}
					}
					if (subST.size() < minTrace)
						continue;

					AttributeShiftSequence assD = new AttributeShiftSequence(shiftsForD);

					System.out.println("------------------------------------------------------");
					System.out.println(
							topActivity + " shift " + noShift + " setting " + o.toString() + " " + subST.size());
					System.out.println("------------------------------------------------------");

					// Create new model for variable that is set by dn
					DMNActivityAttributePair dap = new DMNActivityAttributePair(topActivity, o, values, noShift, 1);
					DMNModel newModel = new DMNModel(
							topActivity + "(" + noShift + ")->" + o.toDotString() + "_" + subST.size(), subST, dap);	

					// Train connection between setted attribute and the activity's other activities
					models.add(newModel);
					findRelatedNodes(1, newModel, assD, dap, subST);			
					// Rude model model is you big enough?
					if (newModel.getEdges().size() > 1) {
						System.out.println("Making model for " + topActivity + " deciding " + o.toString()
								+ " for shift " + noShift);
					}
				}
			}
		}
		Set<DMNModel> ms = eliminateDuplicates(models);
		models = ms;
		for(DMNModel m: ms){
			if(m.getEdges().size()>1){
				dots.add(m.toDot());
			}
		}
	}
	
	protected void findRelatedNodes(int level, DMNModel newModel, AttributeShiftSequence assD, DMNActivityAttributePair topDAP,
			Set<Integer> shiftingTraces){
		boolean f = false;
		
		// Check if model already built
		for(Pair<DMNActivity,DMNAttribute> s: newModel.getTrainedModels())
			if(s.getFirst().equals(topDAP.getActivity()) && s.getSecond().equals(topDAP.getAttribute())){
				f=true;
				break;
			}
		
		String space = "";
		for(int i=1;i<level;i++)
			space+= "|_";
		
		// Model for input-output already constructed
		if(!f && topDAP.getTraces().size()>minTrace){									
			List<DMNActivityAttributePair> pairs = new ArrayList<DMNActivityAttributePair>();

			// For all incoming attributes da set by activity dda
			for (DMNAttribute da : topDAP.getActivity().getInputAttributes()) {
				for (XEventClass e : activities.keySet()) {
					if (activities.get(e).getSettedAttributes().contains(da)) {
						Activity dda = activities.get(e);
													
						//System.out.println(space+"--Checking correlation with ||" + dda.toString() + "|| for " + da);

						// The ass of the other activity + attribute
						AttributeShiftSequence ass = new AttributeShiftSequence(
								dda.getShifts(da));

						// Let's find the number of shifts of da (var)-act (dda) that have a higher support than minTrace
						HashMap<Integer, Set<Integer>> noShiftsFrequency = new HashMap<Integer, Set<Integer>>();

						for (Integer i : shiftingTraces) {
							// assD only contains the noShift of the top
							int earliest = assD.getEarliestAppearanceOf(topDAP.getActivity().getEventClass(), i);

							int noShiftsInTrace = ass.getOrderedShiftsOfTraceBefore(i, earliest).size();
							//System.out.println("Earliest: "+earliest+" size: "+noShiftsInTrace);
							for (int sc = 1; sc <= noShiftsInTrace; sc++) {
								if (!noShiftsFrequency.containsKey(sc))
									noShiftsFrequency.put(sc, new HashSet<Integer>());
								noShiftsFrequency.get(sc).add(i);
							}
						}

						// Let's work with these shifts of var-act pair
						HashSet<Integer> shiftsToConsider = new HashSet<Integer>();
						for (Integer noShifts : noShiftsFrequency.keySet())
							if (noShiftsFrequency.get(noShifts).size() > minTrace && noShifts > 0)
								shiftsToConsider.add(noShifts);
						//System.out.println("Shifts to consider: " + shiftsToConsider.toString());

						for (int daShift : shiftsToConsider) {
							if(newModel.areConnected(topDAP, dda, da, daShift))
								continue;
							
							Map<Integer,String> values = new HashMap<Integer,String>();

							for (Integer i : noShiftsFrequency.get(daShift)) {
								int earliest = assD.getEarliestAppearanceOf(topDAP.getActivity().getEventClass(), i);
								ArrayList<Shift> ss = ass.getOrderedShiftsOfTraceBefore(i, earliest);
								values.put(i,getAttributeValAtTracePos(da, i, ss.get(daShift - 1).getPosition()));
							}

							double corr = 0.0;
							
							DMNActivityAttributePair aap = new DMNActivityAttributePair(dda, da,
									values, daShift, Math.round(corr*100.0)/100.0);
							
							PossibleModel pm = new PossibleModel(topDAP, aap, values.keySet());
							
							corr = buildModel(pm,true).getOutcome();
							
							if (corr > minCorr && values.values().size() > 1 
									&& noShiftsFrequency.get(daShift).size() > minTrace) {
								System.out.println(space+
										dda + " (" + daShift + ") sets " + da + " " + noShiftsFrequency.get(daShift).size()
												+ " times, corr with " + topDAP.getAttribute() + " is " + corr);

								aap.setCorr(Math.round(corr*100.0)/100.0);
								pairs.add(aap);
							}
						}
					}
				}
			}
			// Sort pairs according to the size of traces supporting them
			Collections.sort(pairs);
					
			Set<PossibleModel> pm = 
					findPossibleModels(topDAP, new HashSet<PossibleModel>(), pairs);
			
			Set<PossibleModel> toRemove = new HashSet<PossibleModel>();
			for(PossibleModel p: pm)
				for(PossibleModel p2: pm)
					if(!p.equals(p2))
						if(p.contains(p2))
							toRemove.add(p2);	
			pm.removeAll(toRemove);			
			
			List<PossibleModel> list = Lists.newArrayList(pm);
			Collections.sort(list);
			for(PossibleModel pms: list)
				System.out.println("Possible model "+pms);
			
			for (PossibleModel p : list) {
				if (p.getTraces().size() > minTrace) {
					PredictiveModel predModel = buildModel(p,false);
					if (!modelBased || predModel.getOutcome() > minSup) {	
						if(modelBased)
							topDAP.setClassifier(predModel.getPredModel());
						if (p.getTraces().equals(newModel.getTraces())) {
							System.out.println(
									"\n--------------------------------\nExtending model with " + p.getDAAPS());
							addNodes(level, newModel, topDAP, p);
						} else {
							DMNModel m = newModel.clone(newModel.getLabel() + "_" + topDAP.getTraces().size(),
									p.getTraces());
							
							models.add(m);
							newModel.addSubModel(m);

							System.out.println("\n--------------------------------\nExtending model with new model for "
									+ p.getDAAPS());

							m.setSuperModel(newModel);
							m.setLabel(newModel.getLabel() + "_|_" + p.getTraces().size());
							addNodes(level, m, topDAP, p);
						}
					}
				}
			}
		}
	}
	
}
