package org.processmining.pmind.algorithms.autocorrelations;

import java.io.File;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;

import org.apache.commons.math3.linear.BlockRealMatrix;
import org.apache.commons.math3.linear.RealMatrix;
import org.apache.commons.math3.stat.correlation.PearsonsCorrelation;
import org.apache.commons.math3.stat.correlation.SpearmansCorrelation;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.deckfour.xes.out.XesXmlSerializer;
import org.processmining.pmind.algorithms.concept.AttributeShiftSequence;
import org.processmining.pmind.algorithms.concept.Shift;
import org.processmining.pmind.models.Activity;
import org.processmining.pmind.models.DMNAttribute;

import weka.attributeSelection.CorrelationAttributeEval;
import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;

public class CorrelationsCalculator {

	private double minCorr;
	private double minPair;
	
	private XLog log;
	private Collection<DMNAttribute> dmnAttributes;
	private HashMap<XEventClass,Activity> activities;	
	
	public CorrelationsCalculator(XLog log, Collection<DMNAttribute> dmnAttributes,
			HashMap<XEventClass,Activity> activities, double minCorr, double minPair){
		this.log = log;
		this.dmnAttributes = dmnAttributes;
		this.activities = activities;
		this.minCorr = minCorr;
		this.minPair = minPair;
	}
	
	public static double calculateCorrelation(DMNAttribute at, DMNAttribute at2)
			throws Exception {
		ArrayList<ArrayList<String>> traceData = new ArrayList<ArrayList<String>>();
		ArrayList<Attribute> wekaAttributes = new ArrayList<Attribute>();

		HashSet<Integer> corrTraces = new HashSet<Integer>();
		corrTraces.addAll(at.getAppearingTraces());
		corrTraces.retainAll(at2.getAppearingTraces());
		
		int p = 0;
		for(Integer trace: corrTraces){
			ArrayList<String> pair = new ArrayList<String>();
			pair.add(at.getValueInTrace(trace));
			pair.add(at2.getValueInTrace(trace));
			traceData.add(pair);
			p++;
		}
		
		int i = 0;
		for (int noat = 0; noat < 2; noat++) {
			ArrayList<String> as = new ArrayList<String>();
			for (ArrayList<String> a : traceData)
				as.add(a.get(noat));

			Collection<String> levelsSet = HelpFunctions.getLevels(as);
			ArrayList<String> levels = new ArrayList<String>(levelsSet.size());
			levels.addAll(levelsSet);

			Attribute wekaAttribute;
			if (levels.size() > 50)
				return 0.0;
			

			if (levels.size() > 0)
				wekaAttribute = new Attribute(at + "_" + i, levels);
			else
				wekaAttribute = new Attribute(at + "_" + i);
			wekaAttributes.add(wekaAttribute);
			i++;
		}

		Instances dat = new Instances("Rel", wekaAttributes, traceData.size());
		for (int tr = 0; tr < traceData.size(); tr++) {
			Instance instance = new DenseInstance(2);

			for (int att = 0; att < wekaAttributes.size(); att++) {
				Attribute weA = wekaAttributes.get(att);
				if (weA.isNumeric())
					instance.setValue(weA, Double.valueOf(traceData.get(tr).get(att)));
				else
					instance.setValue(weA, traceData.get(tr).get(att));
			}
			dat.add(instance);
		}

		dat.setClassIndex(dat.numAttributes() - 1);

		CorrelationAttributeEval cae = new CorrelationAttributeEval();
		cae.buildEvaluator(dat);
		for (int j = 0; j < dat.numAttributes() - 1; j++) {
			double correlation = cae.evaluateAttribute(j);
			return correlation;

		}
		return 0.0;
	}
	
	public static double calculateCorrelation(ArrayList<String> val1, ArrayList<String> val2, boolean parametric){
		ArrayList<ArrayList<String>> traceData = new ArrayList<ArrayList<String>>();
		ArrayList<Attribute> wekaAttributes = new ArrayList<Attribute>();

		
		int p = 0;
		for(int i=0;i<val1.size();i++){
			ArrayList<String> pair = new ArrayList<String>();
			pair.add(val1.get(i));
			pair.add(val2.get(i));
			traceData.add(pair);
			p++;
		}
		
		int i = 0;
		for (int noat = 0; noat < 2; noat++) {
			ArrayList<String> as = new ArrayList<String>();
			for (ArrayList<String> a : traceData)
				as.add(a.get(noat));

			Collection<String> levelsSet = HelpFunctions.getLevels(as);
			ArrayList<String> levels = new ArrayList<String>(levelsSet.size());
			levels.addAll(levelsSet);

			Attribute wekaAttribute;
			if (levels.size() > 0)
				wekaAttribute = new Attribute("at_"+noat, levels);
			else
				wekaAttribute = new Attribute("at_"+noat);
			wekaAttributes.add(wekaAttribute);
			i++;
		}

		Instances dat = new Instances("Rel", wekaAttributes, traceData.size());
		for (int tr = 0; tr < traceData.size(); tr++) {
			Instance instance = new DenseInstance(2);

			for (int att = 0; att < wekaAttributes.size(); att++) {
				Attribute weA = wekaAttributes.get(att);
				if (weA.isNumeric())
					instance.setValue(weA, Double.valueOf(traceData.get(tr).get(att)));
				else
					instance.setValue(weA, traceData.get(tr).get(att));
			}
			dat.add(instance);
		}

		dat.setClassIndex(dat.numAttributes() - 1);

		CorrelationAttributeEval cae = new CorrelationAttributeEval();

		if(parametric){
			PearsonsCorrelation pc = new PearsonsCorrelation(datToMat(dat));
			try{
				return Math.abs(pc.getCorrelationMatrix().getEntry(0, 1));
			}catch(Exception e){
				e.printStackTrace();
			}
		}else{
			SpearmansCorrelation sc = new SpearmansCorrelation(datToMat(dat));
			try{
				return Math.abs(sc.getCorrelationMatrix().getEntry(0, 1));
			}catch(Exception e){
				e.printStackTrace();
			}
		}		
//		try {
//			cae.buildEvaluator(dat);
//			for (int j = 0; j < dat.numAttributes() - 1; j++) {
//				double correlation = cae.evaluateAttribute(j);
//				return correlation;
//
//			}
//		} catch (Exception e) {
//			e.printStackTrace();
//		}
		return 0.0;
	}
	
	public static RealMatrix datToMat(Instances dat){
		RealMatrix mat = new BlockRealMatrix(dat.size(), 2);
		int row = 0;
		for(Instance i: dat){
			mat.addToEntry(row, 0, (double) i.value(0));
			mat.addToEntry(row, 1, (double) i.value(1));
			row++;
		}		
		return mat;
	}
	
	
	
	public void searchForAutocorrelations() throws Exception{
		System.out.println("Let's go for autocorrelations: ");
		//HashMap<DMNAttribute,AttributeShiftSequence> asses = new HashMap<DMNAttribute,AttributeShiftSequence>();
		for(DMNAttribute at: dmnAttributes){
			System.out.println("\nGetting shifts of "+at);
			
			//Obtain all shifts
			Collection<Shift> shs = new HashSet<Shift>();
			for(XEventClass act: activities.keySet())
				shs.addAll(activities.get(act).getShifts(at));
			AttributeShiftSequence ass = new AttributeShiftSequence(shs);
			
			
			//Store the values of the shifts
			HashMap<Integer,ArrayList<ArrayList<String>>> valuesPerShiftNo = new HashMap<Integer,ArrayList<ArrayList<String>>>();
			for(int t=0;t<log.size();t++){
				ArrayList<String> values = new ArrayList<String>();
				Collection<Shift> ss= ass.getShiftsOfTrace(t);
				for(Shift s:ss){
					values.add(log.get(t).get(s.getPosition()).getAttributes().get(at.toString()).toString());
				}
				int tU = ss.size();
				//tU=1;
				if(!valuesPerShiftNo.containsKey(tU))
					valuesPerShiftNo.put(tU, new ArrayList<ArrayList<String>>());
				valuesPerShiftNo.get(tU).add(values);				
			}
			
			HashSet<Integer> correlationScore = new HashSet<Integer>();
			for(Integer i: valuesPerShiftNo.keySet()){
				if(i>0 && valuesPerShiftNo.get(i).size()>10){
					System.out.println("For ASSes of length "+i+" - "+valuesPerShiftNo.get(i).size());
					for(int ac=1;ac<5;ac++){
						//if(i>ac)
						if(calculateAutoCorrAll(valuesPerShiftNo.get(i),at.toString(),ac))
							correlationScore.add(ac);
					}
				}
			}
			
			for(int t=0;t<log.size();t++){
				XTrace xt = log.get(t);
				ArrayList<Shift> ss= ass.getOrderedShiftsOfTrace(t);
				for(int lag: correlationScore){	
					for(int sc=ss.size()-1;sc>-1;sc--){	
						Shift s = ss.get(sc);
						XEvent e = xt.get(s.getPosition());
	
						if(sc-lag>-1){
							int pV = ss.get(sc-lag).getPosition();
							XEvent pE = xt.get(pV);
							String value = pE.getAttributes().get(at.toString()).toString();
							
							//System.out.println("Adding "+value+" for "+at.toString()+"_"+lag+" "+e.getAttributes().get(at.toString()).getClass());
							
							XAttribute xa = new XAttributeLiteralImpl(at.toString()+"_"+lag, value);
							e.getAttributes().put(at.toString()+"_"+lag, xa);
						}
					}
				}
			}
		}
		XesXmlSerializer xxs = new XesXmlSerializer();
		FileOutputStream fos = new FileOutputStream(new File("C:/Users/johan/Desktop/test_t.xes"));
		xxs.serialize(log, fos);			
	}
	
	public boolean calculateAutoCorrTrace(int noAtt, ArrayList<ArrayList<String>> data, String at, int ac) throws Exception{
		double corrAvg = 0.0;
		
		int t=0;
		for(ArrayList<String> trace: data){			
			ArrayList<ArrayList<String>> traceData = new ArrayList<ArrayList<String>>();
			ArrayList<Attribute> wekaAttributes = new ArrayList<Attribute>();
			
			//Loop all traces and add pairs to data
			for(int i=trace.size()-1;i>0;i--){
				if(i-ac>-1){
					ArrayList<String> pair = new ArrayList<String>();
					pair.add(trace.get(i));
					pair.add(trace.get(i-ac));
					traceData.add(pair);
				}
			}
			
			int i =0;
			for(int noat=0;noat<2;noat++){
				ArrayList<String> as = new ArrayList<String>();
				for(ArrayList<String> a: traceData)
					as.add(a.get(noat));
				
				Collection<String> levelsSet = HelpFunctions.getLevels(as); 
				ArrayList<String> levels = new ArrayList<String>(levelsSet.size());
				levels.addAll(levelsSet);
				
				Attribute wekaAttribute;
				if(levels.size()>0)
					wekaAttribute = new Attribute(at+"_"+i, levels);
				else
					wekaAttribute = new Attribute(at+"_"+i);
				wekaAttributes.add(wekaAttribute);
				i++;
			}
								
			
			Instances dat = new Instances("Rel", wekaAttributes, traceData.size());				
			for(int tr=0;tr<traceData.size();tr++){
				Instance instance = new DenseInstance(2);
			
				for (int att = 0; att < wekaAttributes.size(); att++) {
					Attribute weA = wekaAttributes.get(att);
					if (weA.isNumeric())
						instance.setValue(weA,
							Double.valueOf(traceData.get(tr).get(att)));
					else
						instance.setValue(weA, traceData.get(tr).get(att));
				}
				dat.add(instance);
			}

			
			dat.setClassIndex(dat.numAttributes()-1);
			
			CorrelationAttributeEval cae = new CorrelationAttributeEval();
			//System.out.println("Building evaluator ");
			cae.buildEvaluator(dat);
			for (int j = 0; j < dat.numAttributes() - 1; j++) {
				//String attribute_name = dat.attribute(j).name();
				double correlation = cae.evaluateAttribute(j);
				corrAvg+= correlation;
				//if(correlation>0.1)
				//	System.out.println(attribute_name + " : " + correlation+" "+ac+" for trace "+trace);
			}
			t++;
		}
		System.out.println("Avg corr ("+ac+"): "+(corrAvg/t)+" ("+t+")");
		double corr = corrAvg/t;
		if(corr>minCorr && t>minPair)
			return true;
		else
			return false;
	}
	
	public boolean calculateAutoCorrAll(ArrayList<ArrayList<String>> data, String at, int ac)
			throws Exception {
		ArrayList<ArrayList<String>> traceData = new ArrayList<ArrayList<String>>();
		ArrayList<Attribute> wekaAttributes = new ArrayList<Attribute>();

		int p = 0;
		for (ArrayList<String> trace : data) {
			//for (int i = trace.size() - 1; i > 0; i--) {
			int i = trace.size()-1;
				if (i - ac > -1) {
					ArrayList<String> pair = new ArrayList<String>();
					pair.add(trace.get(i));
					pair.add(trace.get(i - ac));
					traceData.add(pair);
					p++;
				}
			//}
		}

		int i = 0;
		for (int noat = 0; noat < 2; noat++) {
			ArrayList<String> as = new ArrayList<String>();
			for (ArrayList<String> a : traceData)
				as.add(a.get(noat));

			Collection<String> levelsSet = HelpFunctions.getLevels(as);
			ArrayList<String> levels = new ArrayList<String>(levelsSet.size());
			levels.addAll(levelsSet);

			Attribute wekaAttribute;
			if (levels.size() > 50)
				return false;

			if (levels.size() > 0)
				wekaAttribute = new Attribute(at + "_" + i, levels);
			else
				wekaAttribute = new Attribute(at + "_" + i);
			wekaAttributes.add(wekaAttribute);
			i++;
		}

		Instances dat = new Instances("Rel", wekaAttributes, traceData.size());
		for (int tr = 0; tr < traceData.size(); tr++) {
			Instance instance = new DenseInstance(2);

			for (int att = 0; att < wekaAttributes.size(); att++) {
				Attribute weA = wekaAttributes.get(att);
				if (weA.isNumeric())
					instance.setValue(weA, Double.valueOf(traceData.get(tr).get(att)));
				else
					instance.setValue(weA, traceData.get(tr).get(att));
			}
			dat.add(instance);
		}

		dat.setClassIndex(dat.numAttributes() - 1);

		CorrelationAttributeEval cae = new CorrelationAttributeEval();
		//InfoGainAttributeEval iae = new InfoGainAttributeEval(); 
		//System.out.println("Building evaluator ");
		cae.buildEvaluator(dat);
		//iae.buildEvaluator(dat);
		for (int j = 0; j < dat.numAttributes() - 1; j++) {
			String attribute_name = dat.attribute(j).name();
			double correlation = cae.evaluateAttribute(j);
			if (correlation > minCorr && p > minPair*data.size()) {
				System.out.println(attribute_name + " : " + correlation + " " + ac + " (p=" + p + ")");
				return true;
			}
		}
		return false;
	}
	
}
