package org.processmining.pmind.algorithms.concept;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import org.deckfour.xes.classification.XEventClass;

public class AttributeShiftSequence {

	private Map<Integer,HashSet<Shift>> traces;
	
	private Map<XEventClass,Map<XEventClass,Integer>> ePM;
	private Collection<XEventClass> events = new HashSet<XEventClass>();
	
	public AttributeShiftSequence(Collection<Shift> shifts){	
		// Map<TraceNo,ShiftSequence>
		traces = new HashMap<Integer,HashSet<Shift>>();
		for(Shift s: shifts){
			events.add(s.getEventType());
			if(traces.containsKey(s.getTraceNo())){
				traces.get(s.getTraceNo()).add(s);
			}else{
				HashSet<Shift> aL = new HashSet<Shift>();
				aL.add(s);
				traces.put(s.getTraceNo(), aL);
			}
		}		
		
		// Stores preliminary sequence info for later
		ePM = new HashMap<XEventClass,Map<XEventClass,Integer>>();		
		for(XEventClass e: events){
			HashMap<XEventClass,Integer> hm = new HashMap<XEventClass,Integer>();
			for(XEventClass e2: events){
				if(!e.equals(e2))
					hm.put(e2, 0);
			}
			ePM.put(e,hm);
		}		
		// Calculates how many two activities happen before or after another one
		// Stored in ePM
		for(Integer t: traces.keySet())
			calculateOrders(t);			
	}
	
	public Collection<Shift> getShiftsOfTrace(int traceNo){
		if(traces.get(traceNo) != null)
			return traces.get(traceNo);
		else return new HashSet<Shift>();
	}
	
	public ArrayList<Shift> getOrderedShiftsOfTrace(int traceNo){
		ArrayList<Shift> output = new ArrayList<Shift>();
		if(traces.get(traceNo) != null){
			for(Shift s: traces.get(traceNo))
				output.add(s);
			Collections.sort(output);
		}
		return output;
	}
	
	public ArrayList<Shift> getOrderedShiftsOfTraceBefore(int traceNo, int pos){
		ArrayList<Shift> output = new ArrayList<Shift>();
		if(traces.get(traceNo) != null){
			for(Shift s: traces.get(traceNo))
				if(s.getPosition()<pos)
					output.add(s);
			Collections.sort(output);
		}
		return output;
	}
				
	public int getNthAppearanceOf(XEventClass e, int traceNo, int n){
		int position = Integer.MAX_VALUE;	
		int ePosition = 0;
		for(Shift shift: getOrderedShiftsOfTrace(traceNo)){
			if(shift.getEventType().equals(e))
				ePosition++;
			if(ePosition==n)
				return shift.getPosition();
		}
		return position;
	}
	
	public int getEarliestAppearanceOf(XEventClass e, int traceNo){
		int position = Integer.MAX_VALUE;		
		for(Shift shift: traces.get(traceNo)){
			if(shift.getEventType().equals(e))
				if(shift.getPosition()<position)
					position = shift.getPosition();
		}
		return position;
	}
	
	public Shift getEarliestShiftOf(XEventClass e, int traceNo){
		int position = Integer.MAX_VALUE;		
		Shift result = null;
		for(Shift shift: traces.get(traceNo)){
			if(shift.getEventType().equals(e))
				if(shift.getPosition()<position){
					position = shift.getPosition();
					result = shift;
				}
		}
		return result;
	}
	
	public List<Integer> getAllAppearancesBefore(int traceNo, int pos){
		List<Integer> app = new ArrayList<Integer>();
		if(traces.containsKey(traceNo))
			for(Shift shift: traces.get(traceNo))
				if(shift.getPosition()< pos)
					app.add(shift.getPosition());		
		Collections.sort(app);
		return app;
	}
	
	public Collection<Integer> getAllAppearancesBefore(XEventClass e, int traceNo, int pos){
		Collection<Integer> app = new HashSet<Integer>();
		for(Shift shift: traces.get(traceNo))
			if(shift.getEventType().equals(e))
				if(shift.getPosition()<= pos)
					app.add(shift.getPosition());		
		return app;
	}
	
	public boolean isBefore(XEventClass e1, XEventClass e2){
		if(ePM.get(e1).get(e2)>ePM.get(e2).get(e1))
			return true;
		else
			return false;
	}
	
	public Map<Integer,List<Shift>> getShiftingTracesForCardinality(int cardinality){
		HashMap<Integer,List<Shift>> output = new HashMap<Integer,List<Shift>>();	
		for(Integer i: traces.keySet()){
			if(traces.get(i).size()==cardinality){
				ArrayList<Shift> l = new ArrayList<Shift>();
				for(Shift s: traces.get(i))
					l.add(s);				
				Collections.sort(l);
				output.put(i, l);
			}
		}
		return output;
	}
	
	public HashMap<Integer,HashMap<Integer,List<Shift>>> getShiftingTracesPerCardinality(int sr){
		HashMap<Integer,HashMap<Integer,List<Shift>>> traceSizeCounts 
			= new HashMap<Integer,HashMap<Integer,List<Shift>>>();
		for(Integer i: traces.keySet()){
			int size = traces.get(i).size();
			if(!traceSizeCounts.containsKey(size))
				traceSizeCounts.put(size, new HashMap<Integer,List<Shift>>());
			ArrayList<Shift> l = new ArrayList<Shift>();
			for(Shift s: traces.get(i))
				l.add(s);				
			Collections.sort(l);
			traceSizeCounts.get(size).put(i, l);
		}
		Collection<Integer> tR = new HashSet<Integer>();
		for(Integer i: traceSizeCounts.keySet())
			if(traceSizeCounts.get(i).keySet().size()<sr)
				tR.add(i);
		for(Integer i: tR)
			traceSizeCounts.remove(i);
		return traceSizeCounts;
	}
	
	private void calculateOrders(int trace){		
		HashSet<Shift> shifts = traces.get(trace);

		for(XEventClass e: events){
			int minE1 = Integer.MAX_VALUE;
			for(Shift s: shifts){
				if(s.getEventType().equals(e))
					if(s.getPosition()<minE1)
						minE1=s.getPosition();
			}
			for(XEventClass e2: events){
				if(e!=e2){
					int minE2 = Integer.MAX_VALUE;
					for(Shift s: shifts){
						if(s.getEventType().equals(e2))
							if(s.getPosition()<minE2)
								minE2=s.getPosition();
					}
					if(minE1<minE2 && minE1!=Integer.MAX_VALUE && minE2!=Integer.MAX_VALUE){
						//System.out.println(minE1+" - "+minE2);
						if(ePM.containsKey(e)){
							Map<XEventClass,Integer> hm = ePM.get(e);
							hm.put(e2, hm.get(e2)+1);
						}
					}else{
						if(ePM.containsKey(e2)){
							Map<XEventClass,Integer> hm = ePM.get(e2);
							hm.put(e, hm.get(e)+1);
						}
					}	
				}
			}
		}	
	}
	
	public boolean hasSameActivitySequence(AttributeShiftSequence ass, int cardinality){
		Map<Integer,List<Shift>> l1 =ass.getShiftingTracesForCardinality(cardinality);
		Map<Integer,List<Shift>> l2 =getShiftingTracesForCardinality(cardinality);
		
		//System.out.println("Comparing "+l1.size()+" "+l2.size());
		if(l1.size()!=l2.size())
			return false;
		for(Integer i: l1.keySet()){
			if(!l2.containsKey(i))
				return false;
			//System.out.println("Trace "+i+" ");
			for(int sc=0;sc<l1.get(i).size();sc++){
				System.out.println(l1.get(i).get(sc).getEventType()+" vs "+l2.get(i).get(sc).getEventType());
				if(!l1.get(i).get(sc).getEventType().equals(l2.get(i).get(sc).getEventType()))
					return false;
			}
		}
		return true;
	}
		
}
