package org.processmining.pmind.algorithms.concept;

import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.processmining.framework.util.Pair;
import org.processmining.pmind.models.Activity;
import org.processmining.pmind.models.DMNActivityAttributePair;
import org.processmining.pmind.models.DMNAttribute;

import com.google.common.collect.Lists;

public class PossibleModel implements Comparable<PossibleModel>{

	private Set<DMNActivityAttributePair> daaps = new HashSet<DMNActivityAttributePair>();
	private DMNActivityAttributePair topDAAP;
	private Set<Integer> traces;
	private Set<DMNAttribute> attributes = new HashSet<DMNAttribute>();
	
	public PossibleModel(Set<DMNActivityAttributePair> daap,
			DMNActivityAttributePair topPair, Set<Integer> traces){
		for(DMNActivityAttributePair a: daap){
			daaps.add(a);
			if(!attributes.contains(a.getAttribute()))
				attributes.add(a.getAttribute());
		}
		this.setTopDAAP(topPair);
		this.setTraces(traces);
	}

	public PossibleModel(DMNActivityAttributePair daap,
			DMNActivityAttributePair topPair, Set<Integer> traces){
		daaps.add(daap);
		this.setTopDAAP(topPair);
		this.setTraces(traces);
	}
	
	public void addDAAP(DMNActivityAttributePair a){
		daaps.add(a);
		if(!attributes.contains(a.getAttribute()))
			attributes.add(a.getAttribute());
	}
	
	public Set<DMNActivityAttributePair> getDAAPS(){
		List<DMNActivityAttributePair> d = Lists.newArrayList(daaps);
		Collections.sort(d);
		return new HashSet<DMNActivityAttributePair>(d);
	}
	
	public Set<Pair<Activity,DMNAttribute>> getSingleDAAPS(){
		Set<Pair<Activity,DMNAttribute>> pairs = new HashSet<Pair<Activity,DMNAttribute>>();
		for(DMNActivityAttributePair daap: daaps)
			pairs.add(new Pair<Activity,DMNAttribute>(daap.getActivity(),daap.getAttribute()));
		return pairs;
	}
	
	public String printActivitiesPerAttribute(){
		StringBuilder sb = new StringBuilder();
		for(DMNAttribute a: attributes){
			String output = "\n"+a.toString() + " set by ";
			for(DMNActivityAttributePair daap: daaps){
				if(daap.getAttribute().equals(a))
					output+= daap.getActivity().toString()+'-';
					
			}
			sb.append(output);
		}
		return sb.toString();
	}
	
	public DMNActivityAttributePair getTopDAAP(){
		return topDAAP;
	}
	
	public void setTopDAAP(DMNActivityAttributePair topDAAP){
		this.topDAAP = topDAAP;
	}

	public DMNAttribute getDepVar() {
		return topDAAP.getAttribute();
	}

	public Set<Integer> getTraces() {
		return traces;
	}

	public void setTraces(Set<Integer> traces) {
		this.traces = traces;
	}

	public String toString(){
		return daaps+" -> "+getDepVar()+" ("+traces.size()+")";
	}

	public boolean equals(Object o){
		PossibleModel pm;
		if(o instanceof PossibleModel)
			pm = (PossibleModel) o;
		else
			return false;
		if(pm.getDAAPS().equals(daaps) && pm.getDepVar().equals(getDepVar()) && pm.getTraces().equals(traces))
			return true;
		return false;
	}
	
	public boolean contains(PossibleModel o){
		if(!o.getDepVar().equals(getDepVar()))
			return false;
		if(!o.getTraces().equals(getTraces()))
			return false;
		for(DMNActivityAttributePair daap: o.getDAAPS())
			if(!daaps.contains(daap))
				return false;
		return true;
	}

	public int compareTo(PossibleModel o) {
		if (o.getTraces().size() < getTraces().size())
			return 1;
		else if(o.getTraces().size()==getTraces().size())
			return 0;
		else
			return -1;
	}
	
	
	
}
