package org.processmining.pmind.logs;

import java.awt.FileDialog;
import java.io.File;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;

import javax.swing.JFrame;
import javax.swing.JOptionPane;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.in.XMxmlParser;
import org.deckfour.xes.in.XParser;
import org.deckfour.xes.in.XesXmlParser;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.deckfour.xes.model.impl.XAttributeMapImpl;
import org.deckfour.xes.model.impl.XEventImpl;
import org.deckfour.xes.model.impl.XLogImpl;
import org.deckfour.xes.model.impl.XTraceImpl;
import org.deckfour.xes.out.XesXmlSerializer;
import org.processmining.pmind.algorithms.concept.Shift;
import org.processmining.pmind.models.Activity;

public class EventLogOperations {

	public static void addAttributePositionInformationToLog(XLog log, boolean keepOriginals, int maxAuto){
		XLogInfo xLogInfo = XLogInfoFactory.createLogInfo(log);
		HashSet<String> globalRemovals = new HashSet<String>();
		
		//System.out.println(xLogInfo.getEventAttributeInfo().getAttributeKeys());
		
		for(XTrace t: log){
			HashMap<XEventClass,Integer> eventClasses = new HashMap<XEventClass,Integer>();
									
			for(XEvent e: t){
				XEventClass eC = xLogInfo.getEventClasses().getClassOf(e);
				if(!eventClasses.containsKey(eC))
					eventClasses.put(eC, 0);
				int i = eventClasses.get(eC);
				eventClasses.put(eC, i+1);
				
				HashMap<String,XAttribute> values = new HashMap<String,XAttribute>();
				HashSet<String> removals = new HashSet<String>();
				
				for(String attribute: e.getAttributes().keySet()){
					if(!attribute.equals("concept:name") 
							&& !attribute.equals("time:timestamp") 
							&& !attribute.equals("lifecycle:transition")){ 
						String value = e.getAttributes().get(attribute).toString();						
						
						if(maxAuto==0 || (maxAuto>0 && i<=maxAuto)){
							XAttribute xa = new XAttributeLiteralImpl(attribute+"_"+i, value);
							values.put(attribute+"_"+i, xa);
						}
						removals.add(attribute);
					}
				}
				for(String s: values.keySet())
					e.getAttributes().put(s, values.get(s));
				if(!keepOriginals)
					for(String s: removals)
						e.getAttributes().remove(s);
				globalRemovals.addAll(removals);
			}			
		}	
	}
	
	public static XLog createEventLogOfShifts(XLog log, XLogInfo xLogInfo, 
			XEventClassifier eventClassifier, HashMap<XEventClass,Activity> activities){
		XLog shiftLog = new XLogImpl(log.getAttributes());
		HashSet<Shift> shifts = new HashSet<Shift>();
		for(int t=0;t<log.size();t++){
			int tp = 0;
			XTrace newT = new XTraceImpl(log.get(t).getAttributes());
			newT.getAttributes().put("concept:name", log.get(t).getAttributes().get("concept:name"));
			for(XEvent e: log.get(t)){

				XEvent newE = new XEventImpl();
				XAttributeMap map = new XAttributeMapImpl();
				newE.setAttributes(map);
				
				XEventClass ec = xLogInfo.getEventClasses(eventClassifier).getClassOf(e);
				XEvent previousEvent = null;
				if(tp!=0){	
					previousEvent = log.get(t).get(tp-1);
					//XEventClass pec = xLogInfo.getEventClasses(eventClassifier).getClassOf(e);
				}		
				Collection<Shift> shiftsForE = activities.get(ec).addEvent(e, previousEvent, e, t, tp, xLogInfo);
				shifts.addAll(shiftsForE);				
				tp++;
				
				// Save shifts to file
				ArrayList<String> atts = new ArrayList<String>();
				String actName = e.getAttributes().get("concept:name").toString();
				for(Shift s: shiftsForE)
					atts.add(s.getAttribute().toDotString());
				Collections.sort(atts);
				for(String s: atts)
					actName += "->"+s;
				XAttribute x = new XAttributeLiteralImpl("concept:name",actName);
				if(actName=="")
					newE.getAttributes().put("concept:name", e.getAttributes().get("concept:name"));
				else
					newE.getAttributes().put("concept:name", x);
				if(shiftsForE.size()!=0)
					newT.add(newE);
			}
			shiftLog.add(newT);
		}					
		System.out.println("Shifts found");
		System.out.println("logsize "+ log.size());			
		System.out.println("Finished");
		return shiftLog;
	}

	public static XLog loadEventLog(String file) throws Exception {
		if (file.equals("")) {
			JFrame frame = new JFrame();
			JOptionPane.showMessageDialog(null, "Please select an event log (.mxml/.xes file).");
			FileDialog fd = new FileDialog(frame, "Choose a file", FileDialog.LOAD);
			fd.setDirectory("");
			fd.setFile("*.*");
			fd.setVisible(true);
			frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

			file = fd.getDirectory() + File.separatorChar + fd.getFile();
		}
		File toHandle = new File(file);

		final XLog logIn;
		if (toHandle.getName().contains(".xes")) {
			XParser parser = new XesXmlParser();
			List<XLog> logs = parser.parse(toHandle);
			logIn = logs.get(0);
		} else {
			XParser parser = new XMxmlParser();
			List<XLog> logs = parser.parse(toHandle);
			logIn = logs.get(0);
		}
		return logIn;
	}
	
	public static void writeEventLog(XLog log, String fileLocation) throws Exception {
		XesXmlSerializer xxs = new XesXmlSerializer();
		FileOutputStream fos = new FileOutputStream(new File(fileLocation));
		xxs.serialize(log, fos);	
	}
	
}
