package org.processmining.pmind.models;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.model.XEvent;
import org.processmining.plugins.InductiveMiner.mining.MiningParameters;
import org.processmining.pmind.algorithms.concept.ActivityDataStore;
import org.processmining.pmind.algorithms.concept.Shift;

import weka.core.Attribute;
import weka.core.Instances;

public class Activity {
	boolean VERBOSE = true;
	
	private XEventClassifier eventClassifier = MiningParameters.getDefaultClassifier();
	
	private final XEventClass eventClass;
	//private final boolean isDecisionPoint;
	
	private final Set<DMNAttribute> attributes = new HashSet<DMNAttribute>();
	private final HashMap<DMNAttribute,String> attributeToString = new HashMap<DMNAttribute,String>();
	private final HashMap<String,DMNAttribute> stringToAttribute = new HashMap<String,DMNAttribute>();
	
	private Instances wekaInstances;
	private ArrayList<Attribute> wekaAttributes;
		
	private Instances trainingSet;

	private HashMap<String,ArrayList<String>> values = new HashMap<String,ArrayList<String>>();
	private HashMap<DMNAttribute,Collection<Shift>> shifts = new HashMap<DMNAttribute,Collection<Shift>>();
	
	private Collection<String> indep;
	
	private Collection<DMNAttribute> settingAttributes = new HashSet<DMNAttribute>();
	private Collection<DMNAttribute> inputAttributes = new HashSet<DMNAttribute>();
		
	private Collection<DMNNode> nodes;
	
	private ActivityDataStore ads;
	
	public Activity(XEventClass eventClass, Set<DMNAttribute> attributes, 
			boolean ds, boolean printModels){
		this.eventClass = eventClass;
		this.VERBOSE = printModels;
		
		//isDecisionPoint = ds;
		if(ds)
			System.out.println(eventClass+" is a decisionPoint!");
		for(DMNAttribute att: attributes){
			if(
					!att.toString().equals("concept:name") 
					&& !att.toString().equals("Activity")
				//	|| (att.toString().equals("concept:name") && isDecisionPoint))
					//&& !att.toString().equals("Activity") 
					&& 
					!att.toString().equals("time:timestamp")
					//&& !att.toString().equals("org:resource")
					&& !att.toString().equals("EventID")
					&& !att.toString().equals("lifecycle:transition")){
				this.attributes.add(att);
				attributeToString.put(att, att.toString());
				stringToAttribute.put(att.toString(), att);
				values.put(att.toString(), new ArrayList<String>());
				shifts.put(att, new HashSet<Shift>());
			}
		}
		ads = new ActivityDataStore(attributes);
		nodes = new HashSet<DMNNode>();
	}
	
	public Collection<DMNNode> getNodes(){
		return nodes;
	}
	
	public String getName(){
		return eventClass.toString();//.getId().replace(":", "_").replace(" ", "_").replace("+", "_");
	}
	
	public String toString(){
		return getName();
	}
	
	public XEventClass getEventClass(){
		return eventClass;
	}
	
	public Collection<DMNAttribute> getAttributes(){
		return attributes;
	}
	
	public Collection<DMNAttribute> getFeatures(){
		Collection<DMNAttribute> dmns = new HashSet<DMNAttribute>();
		for(DMNAttribute a: attributes)
			if(indep.contains(a.toString()))
				dmns.add(a);
		return dmns;
	}
	
	public Collection<DMNAttribute> getSettedAttributes(){
		return settingAttributes;
	}	
	
	public Collection<DMNAttribute> getInputAttributes(){
		return inputAttributes;
	}
	
	public void printInstances(){
		System.out.println(wekaInstances.toString());
	}
	
	public List<Shift> getShifts(DMNAttribute a, Integer trace){
		List<Shift> in = new ArrayList<Shift>();
		if(shifts.containsKey(a))
			for(Shift s: shifts.get(a))
				if(s.getTraceNo()==trace)
					in.add(s);	
		Collections.sort(in);
		return in;
	}
	
	public Set<Integer> getShiftTraces(DMNAttribute a){
		HashSet<Integer> in = new HashSet<Integer>();
		for(Shift s: shifts.get(a))
			in.add(s.getTraceNo());
		return in;
	}
	
	public Collection<Shift> getShifts(DMNAttribute a){
		HashSet<Shift> in = new HashSet<Shift>();
		if(shifts.containsKey(a))
			for(Shift s: shifts.get(a))
				in.add(s);	
		return in;
	}
	
	public Collection<Shift> getNthShifts(DMNAttribute a, int n){
		HashSet<Shift> in = new HashSet<Shift>();
		Collection<Integer> traces = getShiftTraces(a);
		for(Integer trace: traces){
			List<Shift> shifts = getShifts(a, trace);
			if(shifts.size()>=n)
				in.add(shifts.get(n-1));
		}
		return in;
	}
		
	public void printShifts(){
		StringBuilder sb = new StringBuilder();
		for(DMNAttribute att: shifts.keySet())
			System.out.println(att+" has "+shifts.get(att).size()+" shifts");
		System.out.println(sb.toString());
	}
		
	// Adds all the information of events to the activity
	// Calculates the shifts based on the difference in value of an attribute between 2 events
	public Collection<Shift> addEvent(XEvent e, XEvent previousEvent, XEvent nextEvent, int t, int tp, XLogInfo info) {
		Collection<Shift> shiftSet = new HashSet<Shift>();

		for (DMNAttribute a : attributes) {
			//Avoid adding missing values
			String attString = a.toString();
			if (e.getAttributes().keySet().contains(attString)) {
				String newValue = e.getAttributes().get(attString).toString();
				String previousValue = "";
				String nextValue = "";
				if (previousEvent != null && previousEvent.getAttributes().get(attString) != null)
					previousValue = previousEvent.getAttributes().get(attString).toString();
				if (nextEvent != null && nextEvent.getAttributes().get(attString) != null)
					nextValue = nextEvent.getAttributes().get(attString).toString();

				if ((!previousValue.equals(nextValue)
						&&tp != 0) || tp==0) {
					Shift s = new Shift(info.getEventClasses(eventClassifier).getClassOf(e),a,t,tp);
					shifts.get(a).add(s);
					shiftSet.add(s);
					//System.out.println("Shift for "+a+" by "+s.getEventType().toString()+" value: "+newValue);
				}

				if (attString.equals("concept:name")) {
					if (nextValue != "") {
						values.get(attString).add(nextValue);
					} else
						values.get(attString).add("");
				} else {
					values.get(attString).add(newValue);
					ads.addDataElement(a, newValue, t);
				}
			} else{
				values.get(attString.toString()).add("");
				ads.addDataElement(a, "", t);
			}
		}
		return shiftSet;
	}
	
	public void analyzeAttributes(double shiftRatio){
		System.out.println("\nActivity "+getName());		
		Collection<String> toRemove = new HashSet<String>();	
		
		// Check whether attribute has more than 1 (shift) value (i.e. is informative)
		for(String attribute: values.keySet()){
			HashSet<String> val = new HashSet<String>();
			for(String ss: values.get(attribute))
				val.add(ss);			
			if(val.size()>1){			
				inputAttributes.add(stringToAttribute.get(attribute));
			}else{
				toRemove.add(attribute);
			}
		}
		for(String s: toRemove)
			values.remove(s);	
		
		// Adds all shifting attributes to the settingAttributes set to use 
		// as outputs later (setting or deciding)
		if(values.keySet().size()>0)
			getShiftingAttributes(shiftRatio);
		//for(DMNAttribute a: attributes)
		//	prepareDatasetCube(a);
		System.out.println("Input: "+inputAttributes.toString());
		System.out.println("Setting: "+settingAttributes.toString());
	}
	
	
	public void getShiftingAttributes(double shiftRatio){
		Collection<DMNAttribute> shiftingAttributes = new HashSet<DMNAttribute>();
		
		// All value lists are the same size, take the first
		int size = values.values().iterator().next().size();
		
		// Find the number of shifts requried
		long minShift = Math.round(shiftRatio * size);		
		System.out.println("MinSHIFT: "+minShift);
		
		// Go over all shifts
		for(DMNAttribute att: shifts.keySet()){			
			// Enough shifts happening
			if(shifts.get(att).size()> minShift){
				System.out.println("->>>>"+ att+" passed minShift");
				shiftingAttributes.add(att);
			}
		}		
		// Setting as in, the ones that shifted
		this.settingAttributes.addAll(shiftingAttributes);
	}
	
//	public double trainModel(DMNAttribute att, double minSup, Collection<Integer> traces){
//		// Split into training/test set
//		prepareDataset(traces);
//		double relation = 0.0;
//		try {
//			relation = trainModel(att, minSup);
//		} catch (Exception e) {
//			e.printStackTrace();
//		}
//		return relation;
//	}
//		
//	public void prepareDataset(Collection<Integer> traces){	
//		wekaAttributes = new ArrayList<Attribute>(values.keySet().size());
//		HashMap<String, ArrayList<String>> data = new HashMap<String, ArrayList<String>>();
//		
//		for (String a : values.keySet()) {
//			//Remove irrelevant traces
//			ArrayList<String> subset = new ArrayList<String>();
//			for(int i=0; i<values.get(a).size();i++)
//				if(traces.contains(i))
//					subset.add(values.get(a).get(i));
//			data.put(a, subset);
//			
//			Attribute wekaAttribute;
//			Collection<String> levels = HelpFunctions.getLevels(subset);
//			if (levels.size() > 1) {
//				ArrayList<String> classVal = new ArrayList<String>(levels.size());
//				classVal.addAll(levels);
//				wekaAttribute = new Attribute(a, classVal);
//				wekaAttributes.add(wekaAttribute);
//			} else if(levels.size()==0){
//				wekaAttribute = new Attribute(a);
//				wekaAttributes.add(wekaAttribute);
//			}
//		}
//		int size = data.values().iterator().next().size();
//		
//		trainingSet = new Instances("Rel", wekaAttributes, size);
//		// Loop all instances
//		for (int i = 0; i < size; i++) {
//			Instance instance = new DenseInstance(data.keySet().size());
//			// Loop all attributes
//			for (int att = 0; att < wekaAttributes.size(); att++) {
//				Attribute weA = wekaAttributes.get(att);
//				if (weA.isNumeric())
//					instance.setValue(weA,
//							Double.valueOf(data.get(weA.name()).get(i)));
//				else
//					instance.setValue(weA, data.get(weA.name()).get(i));
//			}
//			trainingSet.add(instance);
//		}
//	}
//			
//	private double trainModel(DMNAttribute target, double minSup) throws Exception{		
//		int classIndex = 0;
//		
//		// Find index of target variable
//		for(int wa=0;wa<wekaAttributes.size();wa++){
//			if(wekaAttributes.get(wa).name().equals(target.toString())){
//				classIndex = wa;
//				break;
//			}
//		}
//				
//		//Discretize if necessary
//		if(trainingSet.attribute(classIndex).isNumeric()){
//			if(VERBOSE)	
//				System.out.println("Target: "+target+" classIndex "+classIndex);
//			Discretize d = new Discretize(Integer.toString(classIndex+1));
//			d.setInputFormat(trainingSet);					
//			trainingSet = Filter.useFilter(trainingSet, d);
//		}
//		
//		trainingSet.setClassIndex(classIndex);
//		
//		System.out.println("Building a weka model with "+trainingSet.size()+" and "+wekaAttributes.size()+" attributes");
//		
//		
//		//If the attributes have more than two values, and the trainingset is larger than 1, construct model
//		if(!(trainingSet.numDistinctValues(classIndex)==1) && trainingSet.size()>10
//				//&&	trainingSet.testDistinctValues(classIndex)
//				){			
//			//DMNDecisionTable dt = new DMNDecisionTable();
//			//dt.setDisplayRules(true);	
//			//dt.buildClassifier(trainingSet);					
//		
//			Classifier rf = new BayesNet();//J48();
//			if(trainingSet.attribute(classIndex).isNumeric()){
//				System.out.println("MLP");	
//				rf = new SimpleLinearRegression();
//			}
//			
//			rf.buildClassifier(trainingSet);
//			Evaluation ev = new Evaluation(trainingSet);
//			ev.crossValidateModel(rf, trainingSet, 5, new Random(1));
//			
//			indep = new HashSet<String>();
//			for(int i=0;i<wekaAttributes.size();i++)
//				indep.add(trainingSet.attribute(i).name());
//			indep.remove(target.toString());
//			
//			double auc = 0.0;
//			try {
//				auc = ev.areaUnderROC(classIndex);
//			} catch (Exception e) {
//				return 0.0;
//			}
//			System.gc();
//			ev = null;
//			rf = null;
//			
//			if(auc>minSup){
//				//System.out.println("\nRF werkt "+auc+" "+getName());
//				//System.out.println(trainingSet.attribute(classIndex).name());
//				for(String a: indep)
//					System.out.print("  "+a);
//				if(VERBOSE){	
//					System.out.println("\tLabel: "+target);
//					System.out.println("\tInput variables: "+getFeatures());
//				}		
//				return auc;				
//			}else{
//				if(VERBOSE)
//					System.out.println("\t#####Model has no predictive power#####\n");
//			}
//		}else{
//			if(VERBOSE)
//				System.out.println("\t#####Only one value for attribute#####\n");
//		}
//		return 0.0;
//	}		
	
	public void printDependencies(){
		for(DMNAttribute dmnA: this.getSettedAttributes()){
			for(Activity dad: dmnA.getOutputDependencies().keySet()){
				System.out.println(getName()+" sets "+dmnA+" for "+dad.getName());
				System.out.println(getShiftTraces(dmnA).size()+"--"+dad.getShiftTraces(dmnA).size());
				int overlap = 0;
				for(Integer i: getShiftTraces(dmnA))
					if(dad.getShiftTraces(dmnA).contains(i))
						overlap++;
				System.out.println("Overlap: "+overlap);				
			}				
		}
	}
	
//	public void removeCyclicDependencies(){
//		for(DMNAttribute a: this.getSettedAttributes()){
//			connections = new HashSet<Pair<Activity,DMNAttribute>>();
//			if(checkOutputOfAttribute(a, this)){
//				System.out.println("Removing weakest link");
//				removeWeakestLink();
//			}				
//		}
//	}
//	
//	private void removeWeakestLink(){
//		Pair<Activity,DMNAttribute> theChosenOne = null;
//		double lowestScore = Double.MAX_VALUE;
//		for(Pair<Activity,DMNAttribute> p: connections){
//			for(Activity da: p.getRight().getOutputDependencies().keySet()){
//				if(da.equals(p.getLeft()))
//					if(p.getRight().getOutputDependencies().get(da)<lowestScore){
//						theChosenOne = p;
//						lowestScore = theChosenOne.getRight().getOutputDependencies().get(da);
//					}
//			}
//		}
//		System.out.println("The chosen one: "+theChosenOne.getLeft().getName()+" - "+theChosenOne.getRight().toString());
//		theChosenOne.getRight().removeOutputDependency(theChosenOne.getLeft());
//	}
//	
//	private boolean checkOutputOfAttribute(DMNAttribute a, Activity target){
//		System.out.println("Checking: "+a.toString());
//		connections.add(new ImmutablePair<Activity,DMNAttribute>(target, a));
//		for(Activity da: a.getOutputDependencies().keySet()){
//			System.out.println("Checking: "+da.getName());
//			if(da.equals(target)){
//				System.out.println("True!");
//				return true;
//			}else{
//				System.out.println("Down the rabbit hole: "+da.getName());
//				checkOutputOfActivity(da, target);
//			}
//		}		
//		return false;
//	}
//	
//	private boolean checkOutputOfActivity(Activity da, Activity target){
//		for(DMNAttribute a: da.getSettedAttributes()){
//			System.out.println("Down checking: "+a.toString());
//			connections.add(new ImmutablePair<Activity,DMNAttribute>(target,a));
//			if(checkOutputOfAttribute(a, target))
//				return true;
//		}
//		return false;
//	}
	

//	public void prepareDatasetCube(DMNAttribute a){		
//	HashMap<Integer,Collection<Integer>> spt = new HashMap<Integer,Collection<Integer>>();
//	for(Shift s: shifts.get(a)){
//		if(!spt.containsKey(s.getTraceNo()))
//			spt.put(s.getTraceNo(),new HashSet<Integer>());
//		spt.get(s.getTraceNo()).add(s.getPosition());			
//	}
//
//	HashMap<Integer,Integer> diffLengt = new HashMap<Integer,Integer>();
//	for(Integer t: spt.keySet()){
//		if(spt.get(t).size()>0){				
//			if(!diffLengt.containsKey(spt.get(t).size()))
//				diffLengt.put(spt.get(t).size(), 0);
//			diffLengt.put(spt.get(t).size(), diffLengt.get(spt.get(t).size())+1);
//		}
//	}
//					
//	//Different models for different number of shifts in the traces
//	for(Integer i: diffLengt.keySet()){
//		HashMap<String, ArrayList<String>> data = new HashMap<String, ArrayList<String>>();
//		wekaAttributes = new ArrayList<Attribute>();
//
//		if((double) diffLengt.get(i)/(double) spt.keySet().size()>0.2){		
//			//System.out.println("Passed the shift test: "+((double) diffLengt.get(i)/(double) spt.keySet().size()));
//			
//			//Make a model with lag i
//			for (DMNAttribute aa: ads.getAttributes()){
//				ArrayList<String> subset = new ArrayList<String>();
//				//All traces for which the number of shifts corresponds with i
//				for(Integer tr: spt.keySet())
//					if(spt.get(tr).size()==i)
//						subset.add(ads.getValueForAt(aa, tr));
//				data.put(aa.toString(), subset);
//				
//				Attribute wekaAttribute;
//				Collection<String> levels = getLevels(subset);
//				if (levels.size() > 1) {
//					ArrayList<String> classVal = new ArrayList<String>(levels.size());
//					classVal.addAll(levels);
//					wekaAttribute = new Attribute(aa.toString(), classVal);
//					wekaAttributes.add(wekaAttribute);
//				} else if(levels.size()==0){
//					wekaAttribute = new Attribute(aa.toString());
//					wekaAttributes.add(wekaAttribute);
//				}
//			}
//			
//			trainingSet = new Instances("Rel", wekaAttributes, diffLengt.get(i));
//			// Loop all instances -> traces with the right amount of shifts
//			for(int tr=0;tr<diffLengt.get(i);tr++){
//				Instance instance = new DenseInstance(data.keySet().size());
//				// Loop all attributes
//				for (int att = 0; att < wekaAttributes.size(); att++) {
//					Attribute weA = wekaAttributes.get(att);
//					if (weA.isNumeric())
//						instance.setValue(weA,
//								Double.valueOf(data.get(weA.name()).get(tr)));
//					else
//						instance.setValue(weA, data.get(weA.name()).get(tr));
//				}
//				//for(int over=0;over<2;over++)
//					trainingSet.add(instance);
//			}
//			try {
//				double result = trainModel(a, 0.6);
//				if(result>0.0)
//					System.out.println("*****\n\n*****Result! "+ result+" for "+a+" and "+i+" shifts ("+diffLengt.get(i)+")");
//			} catch (Exception e) {
//				e.printStackTrace();
//			}
//		}
//	}
//	
//
//}	
	
	
//	/*
//	 * Get levels of an attribute
//	 */
//	private Collection<String> getLevels(ArrayList<String> data){
//		HashSet<String> levels = new HashSet<String>();
//		boolean numeric = true;
//		for(String s: data){
//			try{
//				Double.valueOf(s);
//			}catch(NumberFormatException f){
//				numeric = false;
//			}
//			levels.add(s);
//		}
//		if(numeric)
//			levels.clear();
//		return levels;
//	}
		
}
