package org.processmining.pmind.models;

import java.util.Map;
import java.util.Set;

import weka.classifiers.Classifier;

public class DMNActivityAttributePair implements Comparable<DMNActivityAttributePair> {
	
	private Activity activity;
	private DMNAttribute attribute;
	private Map<Integer,String> traces;
	private int shift;
	private double corr;
	private Classifier classifier;
	
	public DMNActivityAttributePair(Activity a, DMNAttribute at, Map<Integer,String> t, int i, double corr){
		this.setActivity(a);
		this.setAttribute(at);
		this.traces = t;
		this.setShift(i);
		this.setCorr(corr);
	}
	
	public Set<Integer> getTraces(){
		return traces.keySet();
	}
	
	public String getValueOfTrace(int i){
		return traces.get(i);
	}

	public Activity getActivity() {
		return activity;
	}

	public void setActivity(Activity activity) {
		this.activity = activity;
	}

	public DMNAttribute getAttribute() {
		return attribute;
	}

	public void setAttribute(DMNAttribute attribute) {
		this.attribute = attribute;
	}
	
	public int getShift(){
		return shift;
	}
	
	public void setShift(int s){
		shift = s;
	}
	
	public String toString(){
		return activity+" -> "+attribute.toString()+" ("+traces.size()+","+shift+")";
	}
	
	public double getCorr() {
		return corr;
	}

	public void setCorr(double corr) {
		this.corr = corr;
	}
	
	public int compareTo(DMNActivityAttributePair aap2) {			
		if (getTraces().size() > aap2.getTraces().size())
			return -1;
		else if (getTraces().size() == aap2.getTraces().size())
			return 0;
		else
			return 1;
	}
	
	public boolean weakEquals(Object m){			
		DMNActivityAttributePair daap;
		if(m instanceof DMNActivityAttributePair)
			daap = (DMNActivityAttributePair) m;
		else
			return false;				
		if(daap.getActivity().toString().equals(activity.toString()) 
				&& daap.getAttribute().toString().equals(attribute.toString())
				&& daap.getShift()==shift)
			return true;	
		return false;
	}
	
	public boolean equals(Object m){
		DMNActivityAttributePair daap;
		if(m instanceof DMNActivityAttributePair)
			daap = (DMNActivityAttributePair) m;
		else
			return false;		
		if(daap.getActivity().equals(activity) && daap.getAttribute().equals(attribute)
				&& daap.getShift()==shift && daap.getTraces().equals(traces))
			return true;		
		return false;
	}

	public Classifier getClassifier() {
		return classifier;
	}

	public void setClassifier(Classifier classifier) {
		this.classifier = classifier;
	}
	
}
