package org.processmining.pmind.models;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.processmining.pmind.algorithms.concept.AttributeShiftSequence;

public class DMNAttribute {
	
	private final String name;
	private Map<Activity,Double> inputDependencies = new HashMap<Activity,Double>();
	private Map<Activity,Double> outputDependencies = new HashMap<Activity,Double>();

	private AttributeShiftSequence ass;

	private HashMap<Integer,String> valueInTrace = new HashMap<Integer,String>();
		
	public DMNAttribute(String name){		
		this.name = name;
	}	
	
	public void addASS(AttributeShiftSequence a){
		ass = a;
	}
	
	public AttributeShiftSequence getASS(){
		return ass;
	}
			
	public void addOutputDependencies(Map<Activity,Double> output){
		outputDependencies.putAll(output);
	}
	
	public Map<Activity,Double> getOutputDependencies(){
		return outputDependencies;
	}	
	
	public Set<String> getValues(){
		HashSet<String> values = new HashSet<String>();
		values.addAll(valueInTrace.values());
		return values;
	}
	
	public Collection<Integer> getTraces(){
		return valueInTrace.keySet();
	}
	
	public void addValueOfTrace(int i, String value){
		valueInTrace.put(i, value);
	}
	
	public Collection<Integer> getAppearingTraces(){
		return valueInTrace.keySet();
	}
	
	public String getValueInTrace(int i){
		if(valueInTrace.containsKey(i))
			return valueInTrace.get(i);
		else
			return "";
	}
	
	public String toString(){
		return name;
	}
	
	public String toDotString(){
		return name.replace(":", "_").replace(" ", "_").replace("+", "_").replace("(", "_").replace(")", "_");
	}
	
	public String toVerboseString(){
		StringBuilder sb = new StringBuilder();
		sb.append(name+"\n");
		sb.append("Input: ");
		for(Activity s: inputDependencies.keySet())
			sb.append(s.getName()+" - ");
		sb.append("\n");
		sb.append("Output: ");
		for(Activity s: outputDependencies.keySet())
			sb.append(s.getName()+" - ");
		sb.append("\n");
		return sb.toString();
	}

	@Override
	public boolean equals(Object o1){
		if(o1 instanceof DMNAttribute)
			if(((DMNAttribute) o1).toString().equals(name))
				return true;
		return false;
	}
	
}
