package org.processmining.pmind.models;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.processmining.framework.util.Pair;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.directed.AbstractDirectedGraph;
import org.processmining.models.graphbased.directed.DirectedGraph;
import org.processmining.models.graphbased.directed.DirectedGraphEdge;
import org.processmining.models.graphbased.directed.DirectedGraphElement;
import org.processmining.models.graphbased.directed.DirectedGraphNode;
import org.processmining.plugins.graphviz.dot.Dot;
import org.processmining.plugins.graphviz.dot.Dot.GraphDirection;

public class DMNModel extends AbstractDirectedGraph<DMNNode,DMNEdge> {

	protected Set<DMNNode> nodes = new HashSet<DMNNode>();
	protected Set<DMNEdge> edges = new HashSet<DMNEdge>();
	protected Dot dot = new Dot();
	protected HashSet<Pair<DMNActivity,DMNAttribute>> foundStuff 
		= new HashSet<Pair<DMNActivity,DMNAttribute>>();
	protected Set<DMNModel> submodels = new HashSet<DMNModel>();
	protected DMNModel superModel = null;
	
	protected Set<Integer> traces;
	
	protected Collection<DMNAttribute> attributes = new HashSet<DMNAttribute>();
	protected Collection<Activity> activities = new HashSet<Activity>();
	protected Map<DMNActivityAttributePair,DMNDecisionActivity> aaPairs = new HashMap<DMNActivityAttributePair,DMNDecisionActivity>();
	protected Set<DMNKnowledgeNode> knowledgeNodes = new HashSet<DMNKnowledgeNode>();
	
	protected DMNActivityAttributePair topActivity;
	protected Object[] petrinet;
		
	public DMNModel(String label, Set<Integer> traces, DMNActivityAttributePair tA){
		getAttributeMap().put(AttributeMap.LABEL, label);
		dot.setLabel(label);
		this.traces = traces;

		addNode(tA, "#9D8FFF");
		setTopActivity(tA);
	}	
	
	public void setTopActivity(DMNActivityAttributePair a){
		this.topActivity = a;
	}
		
	public DMNActivityAttributePair getTopActivity(){
		if(!(topActivity==null))
			return topActivity;
		return null;
	}
	
	public int getDepth(DMNActivityAttributePair top, int d){		
		d++;
		int init = d;
		for(DMNActivityAttributePair da: this.getAaPairs()){
			int depth = init;
			if(this.areConnected(da, top, true))
				depth = getDepth(da,depth);
			if(depth>d)
				d = depth;
		}	
		return d;
	}
	
	public DMNDecisionActivity getTopActivityNode(){
		if(!(aaPairs.get(getTopActivity())==null))
			return aaPairs.get(getTopActivity());
		return null;
	}
	
	public DMNDecisionActivity getNode(DMNActivityAttributePair daap){
		return aaPairs.get(daap);
	}
	
	public Dot toDot(){
		dot.setDirection(GraphDirection.bottomTop);	
		return dot;
	}	
	
	public DMNEdge addEdge(DMNNode source, DMNNode target){
		return addEdge(source, target, 1, 0);
	}
	
	public void setLabel(String label){
		getAttributeMap().put(AttributeMap.LABEL, label);
	}
	
	public Set<Integer> getTraces(){
		return traces;
	}
	
	public Set<DMNKnowledgeNode> getKnowledgeNodes(){
		return knowledgeNodes;
	}
	
	public void addSubModel(DMNModel m){
		submodels.add(m);
	}
	
	public Set<DMNModel> getSubModels(){
		return submodels;
	}
		
	public Set<DMNModel> recursivelyGetSubModels(){
		Set<DMNModel> output = new HashSet<DMNModel>();
		for(DMNModel ms: getSubModels()){
			output.add(ms);
			output.addAll(ms.recursivelyGetSubModels());
		}		
		return output;
	}
		
	public void setSuperModel(DMNModel s){
		superModel = s;
	}
	
	public DMNModel getSuperModel(){
		return superModel;
	}
	
	public void addPetrinet(Object[] pn){
		petrinet = pn;
	}
	
	public Object[] getPetrinet(){
		return petrinet;
	}
	
	public DMNKnowledgeNode addKnowledgeNode(String modelDescription, DMNActivityAttributePair daap){
		try{
			aaPairs.get(daap).hashCode();
		}catch(Exception e){
			return null;
		}
		
		DMNKnowledgeNode dkn = new DMNKnowledgeNode(this, dot, daap, "");//modelDescription);
		knowledgeNodes.add(dkn);
		//nodes.add(dkn);
		graphElementAdded(dkn);		
		
		DMNEdge e = new DMNEdge(dkn, aaPairs.get(daap), 0 ,0, dot);
		//edges.add(e);
		graphElementAdded(e);
		//DotEdge ed = dot.addEdge(dkn.getDotNode(), aaPairs.get(daap).getDotNode());
		e.getDotEdge().setLabel("");
		e.getDotEdge().setOption("style", "dashed");			
		return dkn;
	}
	
	public DMNEdge addEdge(DMNNode source, DMNNode target, int weight, double corr){
		checkAddEdge(source, target);
		for(DMNEdge e: edges)
			if(e.getSource().equals(source) && e.getTarget().equals(target)
					&& e.getWeight()==weight && e.getCorr()==corr)
				return e;		
		DMNEdge e = new DMNEdge(source, target, weight, corr, dot);
		edges.add(e);
		graphElementAdded(e);
		
		//DotEdge de = dot.addEdge(e.getSource().getDotNode(), e.getTarget().getDotNode());
		//de.setLabel(String.valueOf(weight)+","+String.valueOf(corr)); 
		return e; 
	}
		
	public DMNNode addNode(DMNActivityAttributePair dap){
		return addNode(dap, "black");
	}
	
	public DMNNode addNode(DMNActivityAttributePair dap, String color){
		for(DMNNode n: nodes)
			if(n instanceof DMNDecisionActivity){			
				DMNDecisionActivity dda = (DMNDecisionActivity) n;
				if(dda.getActivity().equals(dap.getActivity())
						&& dda.getAttribute().equals(dap.getAttribute()) && dda.getShift()==dap.getShift())
					return n;				
			}	
		DMNDecisionActivity dda = new DMNDecisionActivity(this, dot, dap.getActivity(), dap.getAttribute(), dap.getShift(), dap.getTraces());
		
		nodes.add(dda);
		graphElementAdded(dda);
		aaPairs.put(dap,dda);
		
		dot.addNode(dda.getDotNode());
		dda.getDotNode().setOption("shape", "box");
		if(!color.equals("black"))
			dda.getDotNode().setOption("style", "filled");
		dda.getDotNode().setOption("color", color);
		return dda;
	}
	
	// Adding attributes
	public DMNNode addNode(DMNAttribute attr){
		return addNode(attr, "black");
	}
	
	public DMNNode addNode(DMNAttribute attr, String color){
		System.out.println("Adding attribute: "+attr.toString());
		for(DMNNode n: nodes)
			if(n instanceof DMNInput)
				if(((DMNInput) n).getAttribute().equals(attr))
					return n;

		DMNInput n = new DMNInput(this, dot , attr);
		nodes.add(n);
		graphElementAdded(n);
		attributes.add(attr);
		
		dot.addNode(n.getDotNode());
		n.getDotNode().setOption("fillcolor", color);
		return n;
	}
	
	public Set<DMNActivityAttributePair> getInputsOfDAAP(DMNActivityAttributePair daap){
		Set<DMNActivityAttributePair> output = new HashSet<DMNActivityAttributePair>();
		for(DMNActivityAttributePair daap2: getAaPairs())
			if(areConnected(daap2,daap,true))
				output.add(daap2);	
		return output;
	}
	
		
	public boolean areConnected(DMNNode n1, DMNNode n2){
		for(DMNEdge e: this.getEdges())
			if(e.getSource().equals(n1) && e.getTarget().equals(n2)
					|| e.getSource().equals(n2) && e.getTarget().equals(n1))
				return true;
		return false;
	}
	
	public boolean areConnected(DMNActivityAttributePair daap, DMNActivityAttributePair daap2, boolean directed){
		for(DMNEdge e: this.getEdges()){
			DMNDecisionActivity source = null;
			DMNDecisionActivity target = null;
			if(e.getSource() instanceof DMNDecisionActivity)
				source = (DMNDecisionActivity) e.getSource();
			if(e.getTarget() instanceof DMNDecisionActivity)
				target = (DMNDecisionActivity) e.getTarget();
			if(source.equals(aaPairs.get(daap)) && target.equals(aaPairs.get(daap2)))
				return true;
			if(source.equals(aaPairs.get(daap2)) && target.equals(aaPairs.get(daap)) && !directed)
				return true;
		}		
		return false;
	}
	
	public boolean areConnected(DMNDecisionActivity da, DMNDecisionActivity da2){
		for(DMNEdge e: this.getEdges()){
			DMNDecisionActivity source = null;
			DMNDecisionActivity target = null;
			if(e.getSource() instanceof DMNDecisionActivity)
				source = (DMNDecisionActivity) e.getSource();
			if(e.getTarget() instanceof DMNDecisionActivity)
				target = (DMNDecisionActivity) e.getTarget();
			if(source.equals(da) && target.equals(da2))
				return true;
			if(source.equals(da2) && target.equals(da))
				return true;
		}		
		return false;
	}
	
	public DMNEdge getConnection(DMNDecisionActivity da, DMNDecisionActivity da2){
		for(DMNEdge e: this.getEdges()){
			DMNDecisionActivity source = null;
			DMNDecisionActivity target = null;
			if(e.getSource() instanceof DMNDecisionActivity)
				source = (DMNDecisionActivity) e.getSource();
			if(e.getTarget() instanceof DMNDecisionActivity)
				target = (DMNDecisionActivity) e.getTarget();
			if(source.equals(da) && target.equals(da2))
				return e;
			if(source.equals(da2) && target.equals(da))
				return e;
		}		
		return null;
	}
	
	public boolean areConnected(DMNActivityAttributePair daap, Activity a, DMNAttribute at, int shift){
		for(DMNEdge e: this.getEdges()){
			DMNDecisionActivity source = null;
			DMNDecisionActivity target = null;
			if(e.getSource() instanceof DMNDecisionActivity)
				source = (DMNDecisionActivity) e.getSource();
			if(e.getTarget() instanceof DMNDecisionActivity)
				target = (DMNDecisionActivity) e.getTarget();
			if(source.getActivity().equals(a) && source.getAttribute().equals(at) && source.getShift()==shift
					&& target.getActivity().equals(daap.getActivity()) && target.getAttribute().equals(daap.getAttribute())
					&& target.getShift()==daap.getShift())
				return true;
			if(target.getActivity().equals(a) && target.getAttribute().equals(at) && target.getShift()==shift
					&& source.getActivity().equals(daap.getActivity()) && source.getAttribute().equals(daap.getAttribute())
					&& source.getShift()==daap.getShift())
				return true;
		}		
		return false;
	}
	
	public Set<DMNNode> getNodes() {
		return nodes;
	}

	public Set<DMNEdge> getEdges() {
		return edges;
	}

	public Collection<Activity> getActivities(){
		return activities;
	}
	
	public Collection<DMNActivityAttributePair> getAaPairs(){
		return aaPairs.keySet();
	}
	
	public Collection<DMNAttribute> getAttributes(){
		return attributes;
	}
	
	public void removeNode(DirectedGraphNode node) {
		graphElementRemoved(node);		
	}

	@SuppressWarnings("rawtypes")
	public void removeEdge(DirectedGraphEdge edge) {
		graphElementRemoved(edge);		
	}
		
	public boolean equals(Object m){
		DMNModel model;
		if(m instanceof DMNModel)
			model = (DMNModel) m;
		else
			return false;
				
		if(aaPairs.size()!=model.getAaPairs().size())
			return false;
		
		if(attributes.size()!=model.getAttributes().size())
			return false;
		if(edges.size()!=model.getEdges().size())
			return false;
		
		for(DMNActivityAttributePair daap: model.getAaPairs())
			if(!aaPairs.containsKey(daap))
				return false;		
		for(DMNAttribute a: model.getAttributes())
			if(!attributes.contains(a))
				return false;
		
		for(DMNEdge e: model.getEdges()){
			boolean found = false;
			for(DMNEdge e2: getEdges()){
				if(e.getSource().equals(e2.getSource()) 
						&& e.getTarget().equals(e2.getTarget())
						&& e.getWeight()==e2.getWeight()){
					found = true;
					break;
				}
			}
			if(!found)
				return false;
		}
		return true;
	}
	
	public boolean contains(Object m){
		DMNModel model;
		if(m instanceof DMNModel)
			model = (DMNModel) m;
		else
			return false;
				
		for(DMNActivityAttributePair daap: model.getAaPairs()){
			boolean found = false;
			for(DMNActivityAttributePair daap2: aaPairs.keySet())
				if(daap.equals(daap2)){
					found = true;
					break;
				}
			if(!found)
				return false;
		}	
		for(DMNAttribute a: model.getAttributes())
			if(!attributes.contains(a))
				return false;		
		for(DMNEdge e: model.getEdges()){
			boolean found = false;
			for(DMNEdge e2: getEdges()){
				if(e.getSource().equals(e2.getSource()) 
						&& e.getTarget().equals(e2.getTarget())
						&& e.getWeight()<=e2.getWeight()){
					found = true;
					break;
				}
			}
			if(!found)
				return false;
		}
		System.out.println(this+" contains "+m);
		return true;
	}
		
	public DMNModel clone(String label, Set<Integer> traces){
		DMNModel model = new DMNModel(label, traces, topActivity);
		
		for(DMNActivityAttributePair daa: aaPairs.keySet()){
			DMNNode n = model.addNode(daa,aaPairs.get(daa).getDotNode().getOption("color"));
		
			//if(daa.getClassifier()!= null)
			//	model.addKnowledgeNode("", daa);
		}
			
			
		for(DMNAttribute at: attributes)
			model.addNode(at);
		
		for(DMNEdge e: getEdges()){
			DMNNode source= null, target = null;
			for(DMNNode n: model.getNodes()){
				if(n.equals(e.getSource()))
					source = n;
				if(n.equals(e.getTarget()))
					target = n;
			}		
			model.addEdge(source, target, e.getWeight(), e.getCorr());		
		}
		for(Pair<DMNActivity,DMNAttribute> p: foundStuff)
			model.addTrainedModel(p);
		
		//System.out.println("Copied model with "+model.getEdges().size()+" edges");
		return model;
	}
	
	public HashSet<Pair<DMNActivity, DMNAttribute>> getTrainedModels(){
		return foundStuff;
	}
	
	public void addTrainedModel(Pair<DMNActivity,DMNAttribute> p){
		foundStuff.add(p);
	}
	
	protected AbstractDirectedGraph<DMNNode, DMNEdge> getEmptyClone() {
		// TODO Auto-generated method stub
		return null;
	}

	protected Map<? extends DirectedGraphElement, ? extends DirectedGraphElement> cloneFrom(
			DirectedGraph<DMNNode, DMNEdge> graph) {
		// TODO Auto-generated method stub
		return null;
	}

}
