package org.processmining.pmind.parameters;

import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.pmind.parameters.PMInDParameters.MinerType;

import com.fluxicon.slickerbox.components.NiceSlider;
import com.fluxicon.slickerbox.components.NiceSlider.Orientation;
import com.fluxicon.slickerbox.factory.SlickerFactory;

public class PMInDSettingsDialog extends JPanel {

	private static final long serialVersionUID = -60087716353524468L;
	private PMInDParameters parameters = new PMInDParameters();

	public PMInDSettingsDialog(UIPluginContext context) {		
		this.setLayout(new GridLayout(15, 1, 1, 1));

		//Slider 1
		final NiceSlider supportSlider = SlickerFactory.instance().createNiceIntegerSlider("Set minimum support ", 0,
				100, parameters.getMinSup(), Orientation.HORIZONTAL);
		supportSlider.addChangeListener(new ChangeListener() {

			public void stateChanged(ChangeEvent e) {
				parameters.setMinSup(supportSlider.getSlider().getValue());
			}
		});

		//Slider 2
		final NiceSlider shiftSilder = SlickerFactory.instance().createNiceIntegerSlider("Set shift ratio ", 0, 100,
				parameters.getShiftRatio(), Orientation.HORIZONTAL);
		shiftSilder.addChangeListener(new ChangeListener() {

			public void stateChanged(ChangeEvent e) {
				parameters.setShiftRatio(shiftSilder.getSlider().getValue());
			}
		});

		//Slider 3
		final NiceSlider correlationSlider = SlickerFactory.instance().createNiceIntegerSlider(
				"Set minimum correlation ", 0, 100, parameters.getMinCorrelation(), Orientation.HORIZONTAL);
		correlationSlider.addChangeListener(new ChangeListener() {

			public void stateChanged(ChangeEvent e) {
				parameters.setMinCorrelation(correlationSlider.getSlider().getValue());
			}
		});

		//Slider 4
		final NiceSlider deviationSlider = SlickerFactory.instance().createNiceIntegerSlider(
				"Set minimum deviation ratio ", 0, 100, parameters.getMinDeviation(), Orientation.HORIZONTAL);
		deviationSlider.addChangeListener(new ChangeListener() {

			public void stateChanged(ChangeEvent e) {
				parameters.setMinDeviation(deviationSlider.getSlider().getValue());
			}
		});

		//Slider 5
		final NiceSlider traceSlider = SlickerFactory.instance().createNiceIntegerSlider("Set minimum trace ratio ", 0,
				100, parameters.getMinTraceProp(), Orientation.HORIZONTAL);
		traceSlider.addChangeListener(new ChangeListener() {

			public void stateChanged(ChangeEvent e) {
				parameters.setMinTraceProp(traceSlider.getSlider().getValue());
			}
		});
		
		//Slider 5
		final NiceSlider imSlider = SlickerFactory.instance().createNiceIntegerSlider("Set minimum trace ratio ", 0,
				100, parameters.getIMSetting(), Orientation.HORIZONTAL);
		imSlider.addChangeListener(new ChangeListener() {

			public void stateChanged(ChangeEvent e) {
				parameters.setIMSetting(imSlider.getSlider().getValue());
			}
		});

		//ComboBox 1
		@SuppressWarnings("unchecked")
		final JComboBox<MinerType> minerTypes = SlickerFactory.instance().createComboBox(MinerType.values());
		final JCheckBox pmCheckBox = SlickerFactory.instance().createCheckBox("Model-based (vs. only correlation) ", false);
		minerTypes.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				parameters.setMinerType((MinerType) minerTypes.getSelectedItem());
				parameters.setModelBased(pmCheckBox.isSelected());
				if(pmCheckBox.isSelected() && ((MinerType) minerTypes.getSelectedItem()) == MinerType.shiftBased){
					supportSlider.setEnabled(true);
				}else{
					supportSlider.setEnabled(false);
					pmCheckBox.setSelected(false);
					pmCheckBox.setEnabled(false);
				}
			}
		});
		add(minerTypes);

		//Checkbox 1
		pmCheckBox.setSelected(parameters.isModelBased());
		pmCheckBox.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				parameters.setMinerType((MinerType) minerTypes.getSelectedItem());
				parameters.setModelBased(pmCheckBox.isSelected());
				if(pmCheckBox.isSelected() && ((MinerType) minerTypes.getSelectedItem()) == MinerType.shiftBased){
					supportSlider.setEnabled(true);
				}else{
					supportSlider.setEnabled(false);
					pmCheckBox.setSelected(false);
					pmCheckBox.setEnabled(false);
				}
			}
		});
		pmCheckBox.setOpaque(false);
		pmCheckBox.setPreferredSize(new Dimension(100, 30));
		
		//Checkbox 2
		final JCheckBox printCheckBox = SlickerFactory.instance().createCheckBox("Print models per trace", false);
		printCheckBox.setSelected(parameters.isPrintModels());
		printCheckBox.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				parameters.setPrintModels(printCheckBox.isSelected());					
			}
		});
		pmCheckBox.setOpaque(false);
		pmCheckBox.setPreferredSize(new Dimension(100, 30));
		
		add(new JLabel("Minimum support for predictive models"));
		add(supportSlider);
		add(new JLabel("Minimum proportion of shifts necessary"));
		add(shiftSilder);
		add(new JLabel("Minimum correlation between activity-attribute pairs"));
		add(correlationSlider);
		add(new JLabel("Minimum overlap of trace clusters"));
		add(deviationSlider);
		add(new JLabel("Minimum proportion of traces supporting correlation"));
		add(traceSlider);
		add(new JLabel("Noise threshold setting of Inductive Miner"));
		add(imSlider);

		add(printCheckBox);
		add(pmCheckBox);
	}

	public PMInDParameters getParameters() {
		return parameters;
	}

}
