package org.processmining.pmind.plugins;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.Toolkit;

import javax.swing.BoxLayout;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.JTable;

import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.plugins.graphviz.visualisation.DotPanel;
import org.processmining.pmind.algorithms.PMInDModelChecker;
import org.processmining.pmind.algorithms.concept.ModelComparison;
import org.processmining.pmind.help.PMInDHelp;
import org.processmining.pmind.models.DMNActivityAttributePair;
import org.processmining.pmind.models.DMNDecisionActivity;
import org.processmining.pmind.models.DMNEdge;
import org.processmining.pmind.models.DMNModel;
import org.processmining.pmind.models.DMNNode;

public class PMInDCheckerVisualizer {
	
	int scale = 5;
	UIPluginContext context;
	
	Font mainFont;
	Font smallFont = new Font("Helvetica", Font.PLAIN, 15);
	Font bigFont = new Font("Helvetica", Font.PLAIN, 30);
	
	String [] columns = new String[]{"Model","Replay score"};
	JTabbedPane jtp2;

	
	@Plugin(name = "P-MInD Decision Model model checker Visualization",
			parameterLabels = { "P-MInD conformance output"}, 
			returnLabels = { "DMN Models" }, 
			returnTypes = { JComponent.class }, 
			userAccessible = true,
			help = PMInDHelp.TEXT)
	
	@Visualizer
	@UITopiaVariant(affiliation = "University of Edinburgh",
			author = "Johannes De Smedt", 
			email = "johannes.desmedt@ed.ac.uk")
	public JComponent run(UIPluginContext context, PMInDModelChecker input) throws Exception {			
		this.context = context;
		boolean largeScreen = Toolkit.getDefaultToolkit().getScreenSize().getHeight()>1080;
				
		if(largeScreen)
			scale = scale*2;
		
		final JPanel comp = new JPanel();
		comp.setLayout(new BorderLayout());

		JTabbedPane jtp = new JTabbedPane();
		if(largeScreen)
			mainFont = bigFont;
		else
			mainFont = smallFont;

		jtp.setFont(mainFont);
		
		for(ModelComparison m: input.getComparisons()){
			jtp2 = new JTabbedPane();
			jtp2.setPreferredSize(new Dimension(1000,1200));					
			jtp.add(jtp2, m.getModel1()+" vs. "+m.getModel2());	
			
			addModelsToPane(m);
		}
//		for(DMNModel m: input.getUnmatchedModels()){
//			jtp2 = new JTabbedPane();
//			jtp2.setPreferredSize(new Dimension(1000,1200));					
//			jtp.add(jtp2, m.getLabel());	
//			addModelToPane(m);
//		}
		comp.add(jtp, BorderLayout.PAGE_START);
		return comp;
	}
		
	public double checkModels(DMNModel m, DMNModel m2){
		int coverage = 0;
		double contained = 0;
		
		// Compare nodes
		for(DMNActivityAttributePair daap: m.getAaPairs()){
			coverage++;
			boolean f = false;
			for(DMNActivityAttributePair daap2: m2.getAaPairs()){
				if(daap.weakEquals(daap2)){
					f = true;
					break;
				}
			}
			if(f)
				contained++;
		}		
		// Compare edges
		for(DMNEdge e: m.getEdges()){
			coverage++;
			DMNDecisionActivity s = (DMNDecisionActivity) e.getSource();
			DMNDecisionActivity t = (DMNDecisionActivity) e.getTarget();
			DMNDecisionActivity s2 = null;
			DMNDecisionActivity t2 = null;
			for(DMNNode n: m2.getNodes()){
				if(n.getLabel().equals(s.getLabel()))
					s2 = (DMNDecisionActivity) n;
				if(n.getLabel().equals(t.getLabel()))
					t2 = (DMNDecisionActivity) n;
			}
			if(!m2.areConnected(s2,t2)){
			}else{
				double corrDiff = 1 - Math.abs(e.getCorr()- m2.getConnection(s2, t2).getCorr());
				contained+= corrDiff;
			}
		}
		double outcome = contained / coverage;
		return Math.round(outcome * 100.0) / 100.0;
	}
	
	public void colorMissingElements(DMNModel m, DMNModel m2, String color){
		// Compare nodes
		for (DMNActivityAttributePair daap : m.getAaPairs()) {
			boolean f = false;
			for (DMNActivityAttributePair daap2 : m2.getAaPairs()) {
				if (daap.weakEquals(daap2)) {
					f = true;
					break;
				}
			}
			if (!f) {
				m.getNode(daap).getDotNode().setOption("style", "filled");
				m.getNode(daap).getDotNode().setOption("fillcolor", color);
			}
		}

		// Compare edges
		for (DMNEdge e : m.getEdges()) {
			DMNDecisionActivity s = (DMNDecisionActivity) e.getSource();
			DMNDecisionActivity t = (DMNDecisionActivity) e.getTarget();
			DMNDecisionActivity s2 = null;
			DMNDecisionActivity t2 = null;
			for (DMNNode n : m2.getNodes()) {
				if (n.getLabel().equals(s.getLabel()))
					s2 = (DMNDecisionActivity) n;
				if (n.getLabel().equals(t.getLabel()))
					t2 = (DMNDecisionActivity) n;
			}
			if (!m2.areConnected(s2, t2)) {
				e.getDotEdge().setOption("style", "filled");
				e.getDotEdge().setOption("color", color);
			}
		}
	}
	
	private void addModelsToPane(ModelComparison m) throws Exception{		
		JPanel p = new JPanel();
		p.setLayout(new GridLayout(3,1));
		
		DMNModel colorModel1 = m.getModel1().clone(m.getModel1().getLabel(), m.getModel1().getTraces());
		DMNModel colorModel2 = m.getModel2().clone(m.getModel2().getLabel(), m.getModel2().getTraces());
		
		for(DMNActivityAttributePair daap: m.getMissingInM()){
			colorModel1.getNode(daap).getDotNode().setOption("style", "filled");
			colorModel1.getNode(daap).getDotNode().setOption("fillcolor", "red");
		}	
		for(DMNActivityAttributePair daap: m.getMissingInM2()){
			colorModel2.getNode(daap).getDotNode().setOption("style", "filled");
			colorModel2.getNode(daap).getDotNode().setOption("fillcolor", "#F3BE1A");
		}
		for(DMNEdge e: m.getMissingEdgesM1()){
			for(DMNEdge e2: colorModel1.getEdges())
				if(e.equals(e2)){
					e2.getDotEdge().setOption("style", "filled");
					e2.getDotEdge().setOption("color", "red");
				}
		}
		for(DMNEdge e: m.getMissingEdgesM2()){
			for(DMNEdge e2: colorModel2.getEdges())
				if(e.equals(e2)){
					e2.getDotEdge().setOption("style", "filled");
					e2.getDotEdge().setOption("color", "#F3BE1A");
				}
		}
								
		// Add DMN model 1
		JPanel panelM1 = new JPanel();
		panelM1.setLayout(new BoxLayout(panelM1,BoxLayout.X_AXIS));
		
		JLabel mo1 = new JLabel("Model 1");
		mo1.setFont(mainFont);
		panelM1.add(mo1);
		
		DotPanel dotm = new DotPanel(colorModel1.toDot());		
		panelM1.add(dotm);
		
		p.add(panelM1);
		
		// Add DMN model 2
		JPanel panelM2 = new JPanel();
		panelM2.setLayout(new BoxLayout(panelM2,BoxLayout.X_AXIS));
		
		JLabel mo2 = new JLabel("Model 2");
		mo2.setFont(mainFont);
		panelM2.add(mo2);
		
		DotPanel dotm2 = new DotPanel(colorModel2.toDot());		
		panelM2.add(dotm2);
				
		p.add(panelM2);
		
		// Info
		JPanel panelInfo = new JPanel();
		panelInfo.setLayout(new BoxLayout(panelInfo,BoxLayout.X_AXIS));
		
		
		JLabel coverage = new JLabel("Coverage "+m.getScore());
		coverage.setFont(mainFont);
		panelInfo.add(coverage);
		
		// Replay
		String[][] table = new String[m.getReplayResults().size()][2];
		int i = 0;
		for(DMNActivityAttributePair s: m.getReplayResults().keySet()){
			table[i][0] = s.toString();
			table[i][1] = m.getReplayResult(s);
			i++;
		}
		
		JLabel replay = new JLabel("          Replayed models: ");
		replay.setFont(mainFont);
		panelInfo.add(replay);
		
		JTable replayTable = new JTable(table, columns);
		replayTable.setFont(mainFont);
		replayTable.setRowHeight(replayTable.getFontMetrics(mainFont).getHeight());
				
		if(m.getReplayResults().size()>0)
			panelInfo.add(replayTable);
		else
			panelInfo.add(new JLabel("No replay"));
		
		replayTable.doLayout();
		p.add(panelInfo);
		
		// Add everything to tab
		jtp2.add(p, m.getModel1()+" vs. "+m.getModel2());		
		
		for(ModelComparison m2: m.getSubModels())
			addModelsToPane(m2);
	}
	
	private void addModelToPane(DMNModel m){
		JPanel p = new JPanel();
		p.setLayout(new GridLayout(1,1));
		
		// Add DMN model 1
		DotPanel dotm = new DotPanel(m.toDot());
		dotm.setSize((int) dotm.getSize().getWidth()*scale*4, (int) dotm.getSize().getHeight()*scale*4);
		dotm.setPreferredSize(new Dimension((int) dotm.getSize().getWidth()*scale*2,(int) dotm.getSize().getHeight()*scale*2));
		
		p.add(dotm);
		jtp2.add(p, m.getLabel());
	}
			
}
