package org.processmining.pmind.plugins;

import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.pmind.algorithms.AbstractDMNMiner;
import org.processmining.pmind.algorithms.PMInDModelChecker;
import org.processmining.pmind.algorithms.PMInDNoShifts;
import org.processmining.pmind.algorithms.PMInDShiftBased;
import org.processmining.pmind.algorithms.SOAPMInD;
import org.processmining.pmind.help.PMInDHelp;
import org.processmining.pmind.output.PMInDConformanceOutput;
import org.processmining.pmind.output.PMInDOutput;

public class PMInDConformancePlugin {
	
	@Plugin(name = "P-MInD Conformance Checker",
			//parameterLabels = { }, 
			parameterLabels = { "Model for comparison", "Event log"}, 
			returnLabels = { "DMN Models" }, 
			returnTypes = { PMInDModelChecker.class }, 
			userAccessible = true,
			help = PMInDHelp.TEXT)
	
	@UITopiaVariant(affiliation = "University of Edinburgh",
			author = "Johannes De Smedt", 
			email = "johannes.desmedt@ed.ac.uk")
	public PMInDModelChecker run(UIPluginContext context, PMInDOutput input1, XLog log) throws Exception {				
		AbstractDMNMiner miner = null;
		switch(input1.getParameters().getMinerType()){
			case shiftBased: 
				miner = new PMInDShiftBased(log, input1.getParameters());
				break;
			case nonShiftBased:
				miner = new PMInDNoShifts(log, input1.getParameters());
				break;
			case SOAP:
				miner = new SOAPMInD(log, input1.getParameters());
				break;
		}
		miner.mineDMNModels();		

		
		PMInDConformanceOutput o = new PMInDConformanceOutput(miner, input1.getMiner());
		PMInDModelChecker checker = new PMInDModelChecker(o);
		checker.analyzeModels();
						
		return checker;
	}
		
}
