package org.processmining.pmind.plugins;

import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.pmind.algorithms.AbstractDMNMiner;
import org.processmining.pmind.algorithms.PMInDModelChecker;
import org.processmining.pmind.algorithms.PMInDNoShifts;
import org.processmining.pmind.algorithms.PMInDShiftBased;
import org.processmining.pmind.algorithms.SOAPMInD;
import org.processmining.pmind.help.PMInDHelp;
import org.processmining.pmind.logs.EventLogOperations;
import org.processmining.pmind.output.PMInDConformanceOutput;
import org.processmining.pmind.parameters.PMInDParameters;
import org.processmining.pmind.parameters.PMInDSettingsDialog;

public class PMInDConformancePreloaded {
	
	@Plugin(name = "_P-MInD Conformance",
			parameterLabels = { }, 
	//		parameterLabels = { "Event log"}, 
			returnLabels = { "DMN Models" }, 
			returnTypes = { PMInDModelChecker.class }, 
			userAccessible = false,
			help = PMInDHelp.TEXT)
	
	@UITopiaVariant(affiliation = "University of Edinburgh",
			author = "Johannes De Smedt", 
			email = "johannes.desmedt@ed.ac.uk")
	public PMInDModelChecker run(UIPluginContext context/*, XLog log*/) throws Exception {
		PMInDSettingsDialog dialog = new PMInDSettingsDialog(context);
		
		PMInDParameters para2 = new PMInDParameters();
		para2.setModelBased(false);
		para2.setIMSetting(Integer.MAX_VALUE);
		
		//InteractionResult result = context.showWizard("Mine using P-MInD", true, true, dialog);

		XLog log1 = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/BPI17_FeMa16.xes");
		XLog log2 = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/BPI17_JuDe16.xes");
		//log2 = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/17Mini10.xes");
				
		AbstractDMNMiner miner1 = null;
		AbstractDMNMiner miner2 = null;
		switch(dialog.getParameters().getMinerType()){
			case shiftBased: 
				miner1 = new PMInDShiftBased(log1, dialog.getParameters());
				miner2 = new PMInDShiftBased(log2, dialog.getParameters());
				break;
			case nonShiftBased:
				miner1 = new PMInDNoShifts(log1, dialog.getParameters());
				miner2 = new PMInDNoShifts(log2, dialog.getParameters());
				break;
			case SOAP:
				miner1 = new SOAPMInD(log1, dialog.getParameters());
				miner2 = new SOAPMInD(log2, dialog.getParameters());
				break;
		}
		miner1.mineDMNModels();		
		miner2.mineDMNModels();
		
		PMInDConformanceOutput o = new PMInDConformanceOutput(miner1, miner2);
		PMInDModelChecker checker = new PMInDModelChecker(o);
		checker.analyzeModels();
						
		return checker;
	}
		
}
