package org.processmining.pmind.plugins;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.pmind.algorithms.AbstractDMNMiner;
import org.processmining.pmind.algorithms.PMInDNoShifts;
import org.processmining.pmind.algorithms.PMInDShiftBased;
import org.processmining.pmind.algorithms.SOAPMInD;
import org.processmining.pmind.help.PMInDHelp;
import org.processmining.pmind.output.PMInDOutput;
import org.processmining.pmind.parameters.PMInDSettingsDialog;

public class PMInDProMPlugin {
	
	@Plugin(name = "P-MInD Decision Model Mining Framework",
			parameterLabels = { "Event log"}, 
			returnLabels = { "DMN Models" }, 
			returnTypes = { PMInDOutput.class }, 
			userAccessible = true,
			help = PMInDHelp.TEXT)
	
	@UITopiaVariant(affiliation = "University of Edinburgh",
			author = "Johannes De Smedt", 
			email = "johannes.desmedt@ed.ac.uk")
	public PMInDOutput run(UIPluginContext context, XLog log) throws Exception {
		PMInDSettingsDialog dialog = new PMInDSettingsDialog(context);
		
		InteractionResult result = context.showWizard("Mine using P-MInD", true, true, dialog);			
		if(result == InteractionResult.CANCEL){
			context.getFutureResult(0).cancel(false);
			return null;
		}
		
		long beginTime = System.currentTimeMillis();
		
		AbstractDMNMiner miner = null;
		switch(dialog.getParameters().getMinerType()){
			case shiftBased: 
				miner = new PMInDShiftBased(log, dialog.getParameters());
				break;
			case nonShiftBased:
				miner = new PMInDNoShifts(log, dialog.getParameters());
				break;
			case SOAP:
				miner = new SOAPMInD(log, dialog.getParameters());
				break;
		}
		miner.mineDMNModels();
		
		System.out.println("Duration: "+((System.currentTimeMillis()-beginTime)));
		
		return new PMInDOutput(miner, dialog.getParameters());
	}
	
}
