package org.processmining.pmind.plugins;

import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.pmind.algorithms.AbstractDMNMiner;
import org.processmining.pmind.algorithms.PMInDNoShifts;
import org.processmining.pmind.algorithms.PMInDShiftBased;
import org.processmining.pmind.algorithms.SOAPMInD;
import org.processmining.pmind.help.PMInDHelp;
import org.processmining.pmind.logs.EventLogOperations;
import org.processmining.pmind.models.DMNModel;
import org.processmining.pmind.output.PMInDOutput;
import org.processmining.pmind.parameters.PMInDSettingsDialog;

public class PMInDProMPluginPreloaded {
	
	@Plugin(name = "_P-MInD Decision Model Mining Framework",
			parameterLabels = { }, 
	//		parameterLabels = { "Event log"}, 
			returnLabels = { "DMN Models" }, 
			returnTypes = { PMInDOutput.class }, 
			userAccessible = false,
			help = PMInDHelp.TEXT)
	
	@UITopiaVariant(affiliation = "University of Edinburgh",
			author = "Johannes De Smedt", 
			email = "johannes.desmedt@ed.ac.uk")
	public PMInDOutput run(UIPluginContext context/*, XLog log*/) throws Exception {
		PMInDSettingsDialog dialog = new PMInDSettingsDialog(context);
				
		//InteractionResult result = context.showWizard("Mine using P-MInD", true, true, dialog);
		
		XLog log = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/17Mini30.xes");
		log = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/BPI Challenge 2017.xes");
		log = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/BPI17_JuDe16.xes");
//		log = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/BPI17_FeMa16.xes");
		

		
		//EventLogOperations.addAttributePositionInformationToLog(log, true, 2);
		
		long beginTime = System.currentTimeMillis();
		
		AbstractDMNMiner miner = null;
		switch(dialog.getParameters().getMinerType()){
			case shiftBased: 
				miner = new PMInDShiftBased(log, dialog.getParameters());
				break;
			case nonShiftBased:
				miner = new PMInDNoShifts(log, dialog.getParameters());
				break;
			case SOAP:
				miner = new SOAPMInD(log, dialog.getParameters());
				break;
		}
		miner.mineDMNModels();
		
		System.out.println("Duration: "+((System.currentTimeMillis()-beginTime)));
		
		int noModels = miner.getModels().size();
		int noNodes = 0;
		int noEdges = 0;
		int noTrained = 0;
		int depth = 0;
		int maxDepth = 0;
		
		for(DMNModel m: miner.getModels()){
			noNodes += m.getNodes().size();
			noEdges += m.getEdges().size();			
			noTrained += m.getKnowledgeNodes().size();
			
			int mDepth = m.getDepth(m.getTopActivity(), 0);
			if(mDepth>maxDepth)
				maxDepth = mDepth;
			depth+= mDepth;
			noTrained += m.getKnowledgeNodes().size();
		}
		if(noModels>0){
			noNodes = noNodes/noModels;
			noEdges = noEdges/noModels;
			noTrained = noEdges/noModels;
			depth = depth/noModels;
		}
		StringBuilder sb = new StringBuilder();
		sb.append(noNodes+","+noEdges+","+noTrained+","+miner.getTraceClusters().size()+"\n");
		System.out.println(sb.toString());
		
		
		return new PMInDOutput(miner, dialog.getParameters());
	}
	
}
