package org.processmining.pmind.plugins;

import java.io.BufferedWriter;
import java.io.FileWriter;

import org.deckfour.xes.model.XLog;
import org.processmining.pmind.algorithms.AbstractDMNMiner;
import org.processmining.pmind.algorithms.PMInDShiftBased;
import org.processmining.pmind.logs.EventLogOperations;
import org.processmining.pmind.models.DMNModel;
import org.processmining.pmind.parameters.PMInDParameters;

public class PMInDStandaloneBenchmark {
	

	public static void main(String[] args) throws Exception {		
				
		String logName = "BPI17_JuDe16";
		
		XLog log = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/17Mini30.xes");
		log = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/"+logName+".xes");
		//log = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/BPI Challenge 2017.xes");
		//log = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/17Mini10.xes");
			
		int[] minsupports = new int[]{/*50,*/60,70};
		int[] shiftratios = new int[]{10,20,30};
		int[] mincorrs = new int[]{10,20,30};
		int[] mindevs = new int[]{/*80,*/90};
		int[] mintraces = new int[]{10,20,30};
		
		BufferedWriter bw = new BufferedWriter(new FileWriter("C:/Users/johan/Desktop/results_"+logName+".csv"));
		bw.append("minsup,shiftratio,mincorr,mindev,mintrace,time,nonodes,noedges,notrained,depth,maxdepth,traceclusters\n");
		int no = 0;
		int total = minsupports.length * shiftratios.length * mincorrs.length * mindevs.length * mintraces.length;
		for(int sup: minsupports){
			for(int shiftratio: shiftratios){
				for(int corr: mincorrs){
					for(int mindev: mindevs){
						for(int mintrace: mintraces){
							no++;
							System.out.println("\n\n*************"
									+ "Starting: "+sup+","+shiftratio+","+corr+","+mindev+","+mindev+" ("+no+"/"+total+")+\n\n");
							
							
							log = EventLogOperations.loadEventLog("C:/Users/johan/Desktop/"+logName+".xes");
							PMInDParameters param = new PMInDParameters();
							param.setMinSup(sup);
							param.setShiftRatio(shiftratio);
							param.setMinCorrelation(corr);
							param.setMinDeviation(mindev);
							param.setMinTraceProp(mintrace);
							AbstractDMNMiner miner = new PMInDShiftBased(log, param);
							long begin = System.currentTimeMillis();
							miner.mineDMNModels();
							long time = System.currentTimeMillis() - begin;
							
							
							int noModels = miner.getModels().size();
							int noNodes = 0;
							int noEdges = 0;
							int noTrained = 0;
							int depth = 0;
							int maxDepth = 0;
							
							for(DMNModel m: miner.getModels()){
								noNodes += m.getNodes().size();
								noEdges += m.getEdges().size();			
								noTrained += m.getKnowledgeNodes().size();
								
								int mDepth = m.getDepth(m.getTopActivity(), 0);
								if(mDepth>maxDepth)
									maxDepth = mDepth;
								depth+= mDepth;
							}
							if(noModels>0){
								noNodes = noNodes/noModels;
								noEdges = noEdges/noModels;
								noTrained = noEdges/noModels;
								depth = depth/noModels;
							}									
							bw.append(sup+","+shiftratio+","+corr+","+mindev+","+mintrace+","+time+","
									+noNodes+","+noEdges+","+noTrained+","+depth+","+maxDepth+","+miner.getTraceClusters().size()+"\n");
						}
					}
				}
			}
		}
		bw.flush();
		bw.close();
	}
	
}
